from collections.abc import Sequence
from typing import Literal

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


class EnvBaseBucket(s3.Bucket, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        bucket_name: str | None,
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.RETAIN,
        account_id: str = cdk.Aws.ACCOUNT_ID,
        region: str = cdk.Aws.REGION,
        lifecycle_rules: Sequence[s3.LifecycleRule] | None = None,
        inventories: Sequence[s3.Inventory] | None = None,
        auto_delete_objects: bool = False,
        bucket_key_enabled: bool = False,
        block_public_access: s3.BlockPublicAccess | None = s3.BlockPublicAccess.BLOCK_ALL,
        public_read_access: bool = False,
        **kwargs,
    ):
        self.env_base = env_base
        self._full_bucket_name = bucket_name
        if bucket_name is not None:
            self._full_bucket_name = env_base.get_bucket_name(
                base_name=bucket_name, account_id=account_id, region=region
            )
        super().__init__(
            scope,
            id,
            access_control=s3.BucketAccessControl.PRIVATE,
            auto_delete_objects=auto_delete_objects,
            block_public_access=block_public_access,
            bucket_key_enabled=bucket_key_enabled,
            bucket_name=self.bucket_name,
            public_read_access=public_read_access,
            removal_policy=removal_policy,
            lifecycle_rules=lifecycle_rules,
            inventories=inventories,
            **kwargs,
        )

    @property
    def bucket_name(self) -> str:
        return self._full_bucket_name or super().bucket_name

    def grant_permissions(
        self,
        role: iam.IRole | None,
        *permissions: Literal["rw", "r", "w", "d"],
        objects_key_pattern: str | None = None,
    ):
        """Grant Bucket access (r,w,d) to a role, optionally specifying a key pattern

        Args:
            role (iam.IRole | None): role to grant access to
            objects_key_pattern (Optional[str], optional): Optional pattern to constrain access to.
                The pattern is applied to object keys within the bucket. You can use '*' and '?'
                wildcards. For more information, see the following link:
                https://docs.aws.amazon.com/AmazonS3/latest/userguide/security_iam_service-with-iam.html#security_iam_service-with-iam-id-based-policies-resources # noqa: E501
        """  # noqa: E501

        grant_bucket_access(self, role, *permissions, objects_key_pattern=objects_key_pattern)


def grant_bucket_access(
    bucket: s3.Bucket | Sequence[s3.Bucket],
    role: iam.IRole | None,
    *permissions: Literal["rw", "r", "w", "d"],
    objects_key_pattern: str | None = None,
):
    """Grant Bucket access (r,w,d) to a role, optionally specifying a key pattern

    Args:
        bucket (s3.Bucket | Sequence[s3.Bucket]): bucket or buckets to grant access to
        role (iam.IRole | None): role to grant access to
        objects_key_pattern (Optional[str], optional): Optional pattern to constrain access to.
            The pattern is applied to object keys within the bucket. You can use '*' and '?'
            wildcards. For more information, see the following link:
            https://docs.aws.amazon.com/AmazonS3/latest/userguide/security_iam_service-with-iam.html#security_iam_service-with-iam-id-based-policies-resources
            Defaults to None (which in turn represents '*').
    """  # noqa: E501
    if not role:
        return
    for bucket in [bucket] if isinstance(bucket, s3.Bucket) else bucket:
        for bucket_permission in permissions:
            if bucket_permission == "rw":
                bucket.grant_read_write(role, objects_key_pattern=objects_key_pattern)
            elif bucket_permission == "r":
                bucket.grant_read(role, objects_key_pattern=objects_key_pattern)
            elif bucket_permission == "w":
                bucket.grant_write(role, objects_key_pattern=objects_key_pattern)
            elif bucket_permission == "d":
                bucket.grant_delete(role, objects_key_pattern=objects_key_pattern)



from aibs_informatics_core.env import EnvBase
from aws_cdk import Duration
from aws_cdk import aws_sns as sns
from constructs import Construct

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.batch.infrastructure import BatchEnvironment
from aibs_informatics_cdk_lib.constructs_.batch.launch_template import CloudWatchConfigBuilder
from aibs_informatics_cdk_lib.constructs_.cw import EnhancedDashboard


class BatchMonitoring(EnvBaseConstruct):
    def __init__(
        self,
        scope: Construct,
        id: str,
        env_base: EnvBase,
        batch_environments: list[BatchEnvironment],
    ) -> None:
        super().__init__(scope, id, env_base)
        self.alarm_topic = sns.Topic(self, self.get_construct_id("batch-alarm-topic"))
        if self.is_test_or_prod:
            sns.Subscription(
                self,
                self.get_construct_id("batch-alarm-subscription"),
                topic=self.alarm_topic,  # type: ignore
                endpoint="marmotdev@alleninstitute.org",
                protocol=sns.SubscriptionProtocol("EMAIL"),
            )

        self.dashboard_tools = EnhancedDashboard(
            self,
            "batch-dashboard",
            env_base=self.env_base,
            dashboard_name=f"{self.env_base}-Batch-Dashboard",
            start="-P1D",
        )

        self.add_ecs_widgets(batch_environments)

    def add_ecs_widgets(self, batch_environments: list[BatchEnvironment]):
        self.dashboard_tools.add_text_widget(f"Elastic Container Service ({self.env_base})", 1)

        for batch_environment in batch_environments:
            self.dashboard_tools.add_text_widget(
                f"Batch Environment {batch_environment.descriptor}", 2
            )
            builder = CloudWatchConfigBuilder(
                self.env_base, batch_environment.descriptor.get_name()
            )
            self.dashboard_tools.add_graphs(
                grouped_metric_configs=builder.get_grouped_graph_metric_configs(),
                namespace=builder.metric_namespace,
                period=Duration.minutes(5),
                alarm_id_discriminator=batch_environment.descriptor.get_name(),
                alarm_topic=self.alarm_topic,
                dimensions={
                    "env_base": self.env_base,
                    "batch_env_name": batch_environment.descriptor.get_name(),
                },
            )



"""Core AWS utility functions for building ARNs.

This module provides functions for constructing AWS ARNs for various services.
"""

from typing import Literal, cast

import aws_cdk as cdk


def build_arn(
    partition: str = "aws",
    service: str | None = None,
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: str | None = None,
    resource_delim: Literal["/", ":"] = ":",
) -> str:
    """Build an AWS ARN string.

    Args:
        partition (str): AWS partition. Defaults to "aws".
        service (Optional[str]): AWS service name. Defaults to "*".
        region (Optional[str]): AWS region. Defaults to current region.
        account (Optional[str]): AWS account ID. Defaults to current account.
        resource_id (Optional[str]): Resource identifier. Defaults to "*".
        resource_type (Optional[str]): Resource type prefix.
        resource_delim (Literal["/", ":"]): Delimiter between type and ID.

    Returns:
        The constructed ARN string.
    """
    service = service or "*"
    region = region if region is not None else cast(str, cdk.Aws.REGION)
    account = account if account is not None else cast(str, cdk.Aws.ACCOUNT_ID)
    resource_id = resource_id or "*"

    root_arn = f"arn:{partition}:{service}:{region}:{account}"
    if resource_type is not None:
        return f"{root_arn}:{resource_type}{resource_delim}{resource_id}"
    else:
        return f"{root_arn}:{resource_id}"


def build_batch_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["compute-environment", "job", "job-definition", "job-queue"]
    | None = None,  # noqa: E501
) -> str:
    """Build an AWS Batch ARN.

    Args:
        region (Optional[str]): AWS region.
        account (Optional[str]): AWS account ID.
        resource_id (Optional[str]): Resource identifier.
        resource_type (Optional[Literal[...]]): Batch resource type.

    Returns:
        The constructed Batch ARN string.
    """
    return build_arn(
        service="batch",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_dynamodb_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["table"] | None = None,
) -> str:
    """Build an AWS DynamoDB ARN.

    Args:
        region (Optional[str]): AWS region.
        account (Optional[str]): AWS account ID.
        resource_id (Optional[str]): Table name or resource identifier.
        resource_type (Optional[Literal["table"]]): DynamoDB resource type.

    Returns:
        The constructed DynamoDB ARN string.
    """
    return build_arn(
        service="dynamodb",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_ecr_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["repository"] | None = None,
) -> str:
    """Build an AWS ECR ARN.

    Args:
        region (Optional[str]): AWS region.
        account (Optional[str]): AWS account ID.
        resource_id (Optional[str]): Repository name or resource identifier.
        resource_type (Optional[Literal["repository"]]): ECR resource type.

    Returns:
        The constructed ECR ARN string.
    """
    return build_arn(
        service="ecr",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim="/",
    )


def build_sfn_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["*", "activity", "execution", "stateMachine"] | None = None,
) -> str:
    """Build an AWS Step Functions ARN.

    Args:
        region (Optional[str]): AWS region.
        account (Optional[str]): AWS account ID.
        resource_id (Optional[str]): State machine name or resource identifier.
        resource_type (Optional[Literal[...]]): Step Functions resource type.

    Returns:
        The constructed Step Functions ARN string.
    """
    return build_arn(
        service="states",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )


def build_lambda_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["function", "event-source-mapping", "layer"] | None = None,
) -> str:
    """Build an AWS Lambda ARN.

    Args:
        region (Optional[str]): AWS region.
        account (Optional[str]): AWS account ID.
        resource_id (Optional[str]): Function name or resource identifier.
        resource_type (Optional[Literal[...]]): Lambda resource type.

    Returns:
        The constructed Lambda ARN string.
    """
    return build_arn(
        service="lambda",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )


def build_s3_arn(
    region: str | None = None,
    account: str | None = None,
    resource_id: str | None = None,
    resource_type: Literal["bucket", "object", "accesspoint", "job"] | None = None,
) -> str:
    """Build an AWS S3 ARN.

    Note:
        S3 bucket and object ARNs do not include region or account.

    Args:
        region (Optional[str]): AWS region (ignored for bucket/object).
        account (Optional[str]): AWS account ID (ignored for bucket/object).
        resource_id (Optional[str]): Bucket name or object path.
        resource_type (Optional[Literal[...]]): S3 resource type.

    Returns:
        The constructed S3 ARN string.
    """
    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/list_amazons3.html#amazons3-resources-for-iam-policies
    # See table above to see why resource type is set to None
    if resource_type in ["bucket", "object"]:
        resource_type = None
        # ARNs for buckets and objects CANNOT have REGION information
        region = ""
        account = ""

    return build_arn(
        service="s3",
        region=region,
        account=account,
        resource_id=resource_id,
        resource_type=resource_type,
        resource_delim=":",
    )



from collections.abc import Mapping, Sequence
from typing import cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


class EnvBaseVpc(ec2.Vpc, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        max_azs: int = 6,
        ip_addresses: ec2.IIpAddresses | ec2.IpAddresses = ec2.IpAddresses.cidr("10.10.0.0/16"),
        subnet_configuration: Sequence[ec2.SubnetConfiguration] | None = None,
        include_public_subnet: bool = True,
        include_private_subnet: bool = True,
        gateway_endpoints: Mapping[str, ec2.GatewayVpcEndpointOptions] | None = None,
        include_default_endpoints: bool = True,
        include_default_interface_endpoints: bool = True,
        flow_logs: Mapping[str, ec2.FlowLogOptions] | None = None,
        include_default_flow_logs: bool = True,
        nat_gateway_provider: ec2.NatProvider | None = None,
        nat_gateways: int | None = None,
        **kwargs,
    ):
        self.env_base = env_base

        subnet_configuration = list(subnet_configuration or [])
        if include_public_subnet and not any(
            [_ for _ in subnet_configuration if _.subnet_type == ec2.SubnetType.PUBLIC]
        ):
            subnet_configuration.append(
                ec2.SubnetConfiguration(subnet_type=ec2.SubnetType.PUBLIC, name="Public")
            )
        if include_private_subnet and not any(
            [
                _
                for _ in subnet_configuration
                if _.subnet_type == ec2.SubnetType.PRIVATE_WITH_EGRESS
            ]
        ):
            subnet_configuration.append(
                ec2.SubnetConfiguration(
                    subnet_type=ec2.SubnetType.PRIVATE_WITH_EGRESS, name="Private"
                )
            )

        gateway_endpoints = dict(gateway_endpoints or {})
        if include_default_endpoints:
            if "s3_endpoint" not in gateway_endpoints:
                gateway_endpoints["s3_endpoint"] = ec2.GatewayVpcEndpointOptions(
                    service=ec2.GatewayVpcEndpointAwsService.S3
                )
            if "dynamodb_endpoint" not in gateway_endpoints:
                gateway_endpoints["dynamodb_endpoint"] = ec2.GatewayVpcEndpointOptions(
                    service=ec2.GatewayVpcEndpointAwsService.DYNAMODB
                )

        flow_logs = dict(flow_logs or {})
        if include_default_flow_logs:
            if "CW_flow_log" not in flow_logs:
                flow_logs["CW_flow_log"] = ec2.FlowLogOptions(
                    destination=ec2.FlowLogDestination.to_cloud_watch_logs(),
                    traffic_type=ec2.FlowLogTrafficType.ALL,
                )

        super().__init__(
            scope,
            id,
            max_azs=max_azs,
            ip_addresses=cast(ec2.IIpAddresses, ip_addresses),
            subnet_configuration=subnet_configuration,
            gateway_endpoints=gateway_endpoints,
            flow_logs=flow_logs,
            nat_gateway_provider=nat_gateway_provider,
            nat_gateways=nat_gateways,
            **kwargs,
        )

        if include_default_interface_endpoints:
            self.add_interface_endpoint(
                "ecr",
                service=ec2.InterfaceVpcEndpointAwsService.ECR,
            )
            self.add_interface_endpoint(
                "ecr_docker",
                service=ec2.InterfaceVpcEndpointAwsService.ECR_DOCKER,
            )

        cdk.CfnOutput(self, "VpcId", value=self.vpc_id)

    def as_reference(self, scope: constructs.Construct, id: str) -> ec2.IVpc:
        """Generates a VPC reference"""
        return ec2.Vpc.from_lookup(scope, id, vpc_id=self.vpc_id)



import os
from pathlib import Path

import constructs
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.models.aws.s3 import S3KeyPrefix, S3Path
from aws_cdk import aws_s3_deployment as s3deployment
from aws_cdk import aws_ssm as ssm

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.s3 import EnvBaseBucket


class SSMTools(EnvBaseConstruct):
    def __init__(self, scope: constructs.Construct, id: str, env_base: EnvBase):
        super().__init__(scope, id, env_base)
        self.CODE_ASSETS_PREFIX = "code_assets"
        self.FILE_ASSETS_PREFIX = "file_assets"

    def upload_asset(
        self,
        asset_name: str,
        asset_source: s3deployment.ISource,
        destination_bucket: EnvBaseBucket,
        destination_key_prefix: S3KeyPrefix | None = None,
        param_name: str | None = None,
    ) -> tuple[ssm.StringParameter, str]:
        param_name_components = [self.CODE_ASSETS_PREFIX, asset_name]
        if destination_key_prefix is None:
            destination_key_prefix = S3KeyPrefix(self.CODE_ASSETS_PREFIX)

        if param_name:
            parameter_name = self.env_base.get_ssm_param_name(param_name)
        else:
            parameter_name = self.env_base.get_ssm_param_name(*param_name_components)

        filename = f"{asset_name}-asset.zip"

        s3deployment.BucketDeployment(
            self,
            "-".join([*param_name_components, "s3-deployment"]),
            sources=[asset_source],
            destination_bucket=destination_bucket,
            destination_key_prefix=destination_key_prefix + "/",
            retain_on_delete=False,
        )

        param = ssm.StringParameter(
            self,
            "-".join([*param_name_components, "ssm-parameter"]),
            # Store the S3 URI location in an SSM parameter
            string_value=S3Path.build(
                bucket_name=destination_bucket.bucket_name,
                key=f"{destination_key_prefix}/{filename}",
                full_validate=False,
            ),
            description=f"Location of asset {asset_name}",
            parameter_name=parameter_name,
        )
        return param, parameter_name

    def upload_file(
        self,
        path: str | Path,
        destination_bucket: EnvBaseBucket,
        destination_key_prefix: str,
    ) -> tuple[ssm.StringParameter, str]:
        if not os.path.isfile(path):
            raise ValueError(
                f"Cannot upload file and create SSM reference for {path}. Must be a valid file"
            )
        filename = os.path.basename(path)
        name = os.path.splitext(filename)[0]
        destination_key_prefix = os.path.join(self.FILE_ASSETS_PREFIX, destination_key_prefix)
        destination_key = os.path.join(destination_key_prefix, filename)
        param_name_components = [
            *destination_key_prefix.split(os.path.sep),
            name,
        ]

        s3deployment.BucketDeployment(
            self,
            "-".join([name, "s3-deployment"]),
            sources=[s3deployment.Source.asset(os.path.dirname(path))],
            destination_bucket=destination_bucket,
            destination_key_prefix=destination_key_prefix,
        )

        # Store the S3 URI location in an SSM parameter
        s3_uri = S3Path.build(destination_bucket.bucket_name, destination_key, full_validate=False)

        parameter_name = self.env_base.get_ssm_param_name(*param_name_components)
        cmd_wrapper_param = ssm.StringParameter(
            self,
            "-".join([name, "ssm-parameter"]),
            string_value=s3_uri,
            description=f"Location of {filename}",
            parameter_name=parameter_name,
        )
        return cmd_wrapper_param, parameter_name



from typing import cast

from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_iam as iam
from constructs import Construct

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.efs.file_system import (
    EnvBaseFileSystem,
    grant_connectable_file_system_access,
)


class DebugInstanceConstruct(EnvBaseConstruct):
    """
    DebugInstanceConstruct is a CDK construct that creates an EC2 instance pre-configured for debugging and troubleshooting purposes within a given VPC. This instance is designed primarily to facilitate runtime inspection, diagnostics, and interactions with attached resources, including optional file system mounts (EFS) for scenarios such as shared storage debugging or configuration verification.

    The construct provisions:
      - A dedicated security group for the instance.
      - An IAM role with necessary policies (including AmazonSSMManagedInstanceCore and AmazonS3ReadOnlyAccess), optionally supplemented by user-specified inline policies.
      - A Linux-based EC2 instance (defaulting to the latest Amazon Linux 2) with user data commands to perform system updates, install EFS utilities, and debugging tools (e.g., jq, tree).

    Parameters:
      - scope (Construct): The scope in which this construct is defined.
      - id (Optional[str]): The unique identifier for the construct.
      - env_base (EnvBase): The environment configuration required for the construct.
      - vpc (ec2.Vpc): The VPC within which the instance is launched.
      - name (str, optional): Base name for resources. Defaults to "DebugInstance".
      - efs_filesystems (Optional[List[Union[efs.IFileSystem, EnvBaseFileSystem]]], optional): A list of EFS file systems to mount on the instance. If provided, each file system is mounted at a dedicated path under /mnt/efs/.
      - instance_type (ec2.InstanceType, optional): The EC2 instance type for the debug instance. Defaults to t3.medium.
      - machine_image (Optional[ec2.IMachineImage], optional): The machine image used for the instance. Defaults to the latest Amazon Linux 2 image.
      - instance_name (Optional[str], optional): A custom name for the EC2 instance. If not provided, a name is generated based on the base name.
      - instance_role_name (Optional[str], optional): The name of the IAM role assigned to the instance. If not provided, a default name derived from the base name is used.
      - instance_role_policy_statements (Optional[List[iam.PolicyStatement]], optional): Additional IAM policy statements to attach inline to the instance role.

    Usage Examples:
        1. Minimal usage:
         >>> instance = DebugInstanceConstruct(
                 scope=app,
                 id="DebugInstance",
                 env_base=env,
                 vpc=my_vpc
             )
         This creates an EC2 instance with default parameters and without mounting any EFS file systems.

        2. Advanced usage with custom instance details and EFS mounting:
         >>> instance = DebugInstanceConstruct(
                 scope=app,
                 id="CustomDebugInstance",
                 env_base=env,
                 vpc=my_vpc,
                 name="CustomDebug",
                 instance_type=ec2.InstanceType("t3.large"),
                 instance_name="CustomDebugEC2",
                 instance_role_policy_statements=[custom_policy_statement],
                 efs_filesystems=[my_efs]
             )
         This creates an EC2 instance with a custom instance type, a specified instance name, an inline policy, and mounts the provided EFS file system.

        3. Using a custom machine image:
         >>> custom_image = ec2.MachineImage.from_lookup(name="MyCustomAMI")
         >>> instance = DebugInstanceConstruct(
                 scope=app,
                 id="CustomImageInstance",
                 env_base=env,
                 vpc=my_vpc,
                 machine_image=custom_image
             )
        This example shows how to specify a custom machine image instead of using the default Amazon Linux 2 image.
    """  # noqa: E501

    def __init__(
        self,
        scope: Construct,
        id: str | None,
        env_base: EnvBase,
        vpc: ec2.Vpc,
        name: str = "DebugInstance",
        efs_filesystems: list[efs.IFileSystem | EnvBaseFileSystem] | None = None,
        instance_type: ec2.InstanceType = ec2.InstanceType("t3.medium"),
        machine_image: ec2.IMachineImage | None = None,
        instance_name: str | None = None,
        instance_role_name: str | None = None,
        instance_role_policy_statements: list[iam.PolicyStatement] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(scope, id, env_base, **kwargs)

        # Security group for the instance
        self.debug_sg = ec2.SecurityGroup(
            self,
            f"{name}-SG",
            vpc=vpc,
            allow_all_outbound=True,
            description="Security group for Debug EC2 Instance",
        )

        # IAM role for the instance
        self.instance_role = iam.Role(
            self,
            f"{name}-InstanceRole",
            assumed_by=cast(iam.IPrincipal, iam.ServicePrincipal("ec2.amazonaws.com")),
            role_name=self.get_resource_name(instance_role_name or f"{name}-InstanceRole"),
            description="Role for Debug EC2 Instance",
            managed_policies=[
                iam.ManagedPolicy.from_aws_managed_policy_name("AmazonSSMManagedInstanceCore"),
                iam.ManagedPolicy.from_aws_managed_policy_name("AmazonS3ReadOnlyAccess"),
            ],
            inline_policies={
                "UserSpecifiedPolicies": iam.PolicyDocument(
                    statements=instance_role_policy_statements,
                )
            }
            if instance_role_policy_statements
            else None,
        )

        if machine_image is None:
            machine_image = ec2.MachineImage.latest_amazon_linux2(
                storage=ec2.AmazonLinuxStorage.GENERAL_PURPOSE,
                cpu_type=ec2.AmazonLinuxCpuType.X86_64,
                edition=ec2.AmazonLinuxEdition.STANDARD,
                user_data=(user_data := ec2.UserData.for_linux()),
            )
            user_data.add_commands(
                "yum -y update",
                # These are necessary for EFS mounting
                "yum -y install amazon-efs-utils",
                # These are useful for debugging
                "yum -y install jq tree",
            )

        # Create the EC2 Instance
        self.instance = ec2.Instance(
            self,
            f"{name}-Instance",
            instance_type=instance_type,
            machine_image=cast(ec2.IMachineImage, machine_image),
            instance_name=self.get_resource_name(instance_name or f"{name}-Instance"),
            vpc=vpc,
            role=cast(iam.IRole, self.instance_role),
            vpc_subnets=ec2.SubnetSelection(subnet_type=ec2.SubnetType.PUBLIC),
            security_group=self.debug_sg,
        )

        # If any EFS file systems are passed in, mount each one
        if efs_filesystems:
            for i, filesystem in enumerate(efs_filesystems):
                # Allow our instance to connect on the EFS mount target port
                filesystem.connections.allow_default_port_from(self.debug_sg)

                # A path for each EFS mount
                if isinstance(filesystem, EnvBaseFileSystem):
                    mount_path = f"/mnt/efs/{filesystem.file_system_name}"
                else:
                    mount_path = f"/mnt/efs/{filesystem.file_system_id}"

                self.instance.user_data.add_commands(
                    f"mkdir -p {mount_path}",
                    f"mount -t efs -o tls {filesystem.file_system_id}:/ {mount_path}",
                )
                grant_connectable_file_system_access(filesystem, self.instance, "rw")



"""Base stage classes for CDK Pipeline deployments.

This module provides stage classes for use with CDK Pipelines,
enabling environment-aware deployment stages.
"""

import aws_cdk as cdk
import constructs

from aibs_informatics_cdk_lib.project.config import StageConfig
from aibs_informatics_cdk_lib.stacks.base import EnvBaseStack, EnvBaseStackMixins


class ConfigBasedStage(cdk.Stage, EnvBaseStackMixins):
    """CDK Stage configured from a StageConfig object.

    Provides environment-aware stage functionality with automatic
    tagging and configuration-driven setup.

    """

    def __init__(
        self, scope: constructs.Construct, id: str | None, config: StageConfig, **kwargs
    ) -> None:
        """Initialize a configuration-based stage.

        Args:
            scope (constructs.Construct): The parent construct scope.
            id (Optional[str]): The stage ID. Auto-generated if None.
            config (StageConfig): The stage configuration object.
            **kwargs: Additional arguments passed to cdk.Stage.
        """
        super().__init__(
            scope, id or config.env.env_base.get_construct_id(self.__class__.__name__), **kwargs
        )
        self.config = config
        self.env_base = config.env.env_base
        self.add_tags(*self.stage_tags)

    def get_stack_name(self, stack_class: type[cdk.Stack] | str, *names: str) -> str:
        """Get a stage-qualified stack name.

        Args:
            stack_class (Union[Type[cdk.Stack], str]): Stack class or name.
            *names (str): Additional name components.

        Returns:
            The fully qualified stack name for this stage.
        """
        return self.env_base.get_stage_name(
            stack_class.__name__ if not isinstance(stack_class, str) else stack_class, *names
        )

    @property
    def stage_tags(self) -> list[cdk.Tag]:
        return [
            *self.construct_tags,
            cdk.Tag(key=self.env_base.ENV_BASE_KEY, value=self.env_base),
        ]

    @property
    def env_base_stacks(self) -> list[EnvBaseStack]:
        return [
            node
            for node in self.node.find_all(constructs.ConstructOrder.PREORDER)
            if isinstance(node, EnvBaseStack)
        ]

    @property
    def env(self) -> cdk.Environment:
        try:
            return self._env  # type: ignore
        except AttributeError:
            self._env = cdk.Environment(
                account=self.config.env.account,
                region=self.config.env.region,
            )
            return self.env

    @env.setter
    def env(self, env: cdk.Environment):
        self._env = env



"""EFS file system constructs and utilities.

This module provides CDK constructs for creating and managing EFS file systems,
access points, and mount point configurations.
"""

import logging
from dataclasses import dataclass
from typing import Any, Literal, TypeVar

import aws_cdk as cdk
import constructs
from aibs_informatics_aws_utils.batch import to_mount_point, to_volume
from aibs_informatics_aws_utils.constants.efs import (
    EFS_ROOT_PATH,
    EFS_SCRATCH_PATH,
    EFS_SHARED_PATH,
    EFS_TMP_PATH,
    EFSTag,
)
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda as lambda_
from aws_cdk.aws_efs import (
    LifecyclePolicy,
    OutOfInfrequentAccessPolicy,
    PerformanceMode,
    ThroughputMode,
)

from aibs_informatics_cdk_lib.common.aws.iam_utils import grant_managed_policies
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct, EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.sfn.utils import convert_to_sfn_api_action_case

logger = logging.getLogger(__name__)

T = TypeVar("T")


class EnvBaseFileSystem(efs.FileSystem, EnvBaseConstructMixins):
    """Environment-aware EFS file system construct.

    Extends the standard EFS FileSystem with environment base naming conventions
    and helper methods for access point creation.

    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        vpc: ec2.IVpc,
        file_system_name: str,
        allow_anonymous_access: bool | None = None,
        enable_automatic_backups: bool | None = None,
        encrypted: bool | None = None,
        lifecycle_policy: LifecyclePolicy | None = None,
        out_of_infrequent_access_policy: OutOfInfrequentAccessPolicy | None = None,
        performance_mode: PerformanceMode | None = None,
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.DESTROY,
        throughput_mode: ThroughputMode | None = ThroughputMode.BURSTING,
        **kwargs,
    ) -> None:
        """Initialize an environment-aware EFS file system.

        Args:
            scope (constructs.Construct): The construct scope.
            id (str): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
            vpc (ec2.IVpc): VPC for the file system.
            file_system_name (str): Name for the file system.
            allow_anonymous_access (Optional[bool]): Allow anonymous access.
            enable_automatic_backups (Optional[bool]): Enable automatic backups.
            encrypted (Optional[bool]): Enable encryption.
            lifecycle_policy (Optional[LifecyclePolicy]): Lifecycle policy.
            out_of_infrequent_access_policy (Optional[OutOfInfrequentAccessPolicy]):
                Policy for moving files out of infrequent access.
            performance_mode (Optional[PerformanceMode]): Performance mode.
            removal_policy (cdk.RemovalPolicy): Removal policy.
                Defaults to DESTROY.
            throughput_mode (Optional[ThroughputMode]): Throughput mode.
                Defaults to BURSTING.
            **kwargs: Additional arguments passed to parent.
        """
        self.env_base = env_base
        super().__init__(
            scope,
            id,
            vpc=vpc,
            file_system_name=(full_file_system_name := self.get_name_with_env(file_system_name)),
            allow_anonymous_access=allow_anonymous_access,
            enable_automatic_backups=enable_automatic_backups,
            encrypted=encrypted,
            lifecycle_policy=lifecycle_policy,
            out_of_infrequent_access_policy=out_of_infrequent_access_policy,
            performance_mode=performance_mode,
            removal_policy=removal_policy,
            throughput_mode=throughput_mode,
            **kwargs,
        )
        self._file_system_name = full_file_system_name

    @property
    def file_system_name(self) -> str:
        """Get the full file system name including environment prefix.

        Returns:
            The file system name.
        """
        return self._file_system_name

    def create_access_point(
        self, name: str, path: str, *tags: EFSTag | tuple[str, str]
    ) -> efs.AccessPoint:
        """Create an EFS access point.

        Uses CfnAccessPoint because the AccessPoint construct does not support
        tagging or naming. Tags are used to set the name.

        Args:
            name (str): Name used for construct ID and as default Name tag.
            path (str): Access point path within the file system.
            *tags (Union[EFSTag, Tuple[str, str]]): Variable number of tags
                to add to the access point.

        Returns:
            The created access point.
        """
        ap_tags = [tag if isinstance(tag, EFSTag) else EFSTag(*tag) for tag in tags]
        if not any(tag.key == "Name" for tag in ap_tags):
            ap_tags.insert(0, EFSTag("Name", name))

        cfn_access_point = efs.CfnAccessPoint(
            self.get_stack_of(self),
            self.get_construct_id(self.node.id, name, "cfn-ap"),
            file_system_id=self.file_system_id,
            access_point_tags=[
                efs.CfnAccessPoint.AccessPointTagProperty(key=tag.key, value=tag.value)
                for tag in ap_tags
            ],
            posix_user=efs.CfnAccessPoint.PosixUserProperty(
                gid="0",
                uid="0",
            ),
            root_directory=efs.CfnAccessPoint.RootDirectoryProperty(
                creation_info=efs.CfnAccessPoint.CreationInfoProperty(
                    owner_gid="0",
                    owner_uid="0",
                    permissions="0777",
                ),
                path=path,
            ),
        )
        return efs.AccessPoint.from_access_point_attributes(
            self,
            self.get_construct_id(name, "access-point"),
            access_point_id=cfn_access_point.attr_access_point_id,
            file_system=self,
        )  # type: ignore

    def as_lambda_file_system(self, access_point: efs.AccessPoint) -> lambda_.FileSystem:
        """Convert to a Lambda file system configuration.

        Args:
            access_point (efs.AccessPoint): The access point to use.

        Returns:
            Lambda FileSystem configured for the access point.
        """
        ap = access_point
        return lambda_.FileSystem.from_efs_access_point(
            ap=ap,
            # Must start with `/mnt` per lambda regex requirements
            mount_path="/mnt/efs",
        )

    def grant_lambda_access(self, resource: lambda_.Function) -> None:
        """Grant a Lambda function access to this file system.

        Args:
            resource (lambda_.Function): The Lambda function to grant access to.
        """
        grant_file_system_access(self, resource)


class EFSEcosystem(EnvBaseConstruct):
    """Complete EFS ecosystem with predefined access points.

    Creates an EFS file system with root, shared, scratch, and tmp access points.

    Attributes:
        root_access_point: Access point for the root directory.
        shared_access_point: Access point for shared data.
        scratch_access_point: Access point for scratch data.
        tmp_access_point: Access point for temporary data.
    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        file_system_name: str,
        vpc: ec2.Vpc,
        efs_lifecycle_policy: efs.LifecyclePolicy | None = None,
    ) -> None:
        """Initialize an EFS ecosystem.

        Args:
            scope (constructs.Construct): The construct scope.
            id (Optional[str]): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
            file_system_name (str): Name for the file system.
            vpc (ec2.Vpc): VPC for the file system.
            efs_lifecycle_policy (Optional[efs.LifecyclePolicy]): Lifecycle policy.

        Note:
            If the EFS filesystem is intended to be deployed in BURSTING throughput mode,
            it may be counterproductive to set an efs_lifecycle_policy other than None
            because EFS files in IA tier DO NOT count towards burst credit accumulation.
            See: https://docs.aws.amazon.com/efs/latest/ug/performance.html#bursting
        """
        super().__init__(scope, id, env_base)
        self._file_system = EnvBaseFileSystem(
            scope=self,
            id=f"{file_system_name}-fs",
            env_base=self.env_base,
            file_system_name=self.get_name_with_env(file_system_name),
            lifecycle_policy=efs_lifecycle_policy,
            out_of_infrequent_access_policy=efs.OutOfInfrequentAccessPolicy.AFTER_1_ACCESS,
            enable_automatic_backups=False,
            throughput_mode=efs.ThroughputMode.BURSTING,
            removal_policy=cdk.RemovalPolicy.DESTROY,
            vpc=vpc,
        )

        self.root_access_point = self.file_system.create_access_point(
            name="root", path=EFS_ROOT_PATH
        )
        self.shared_access_point = self.file_system.create_access_point(
            name="shared", path=EFS_SHARED_PATH
        )
        self.scratch_access_point = self.file_system.create_access_point(
            name="scratch", path=EFS_SCRATCH_PATH
        )
        self.tmp_access_point = self.file_system.create_access_point(name="tmp", path=EFS_TMP_PATH)
        self.file_system.add_tags(cdk.Tag("blah", self.env_base))

    @property
    def file_system(self) -> EnvBaseFileSystem:
        """Get the underlying EFS file system.

        Returns:
            The EFS file system instance.
        """
        return self._file_system

    @property
    def as_lambda_file_system(self) -> lambda_.FileSystem:
        """Get the file system configured for Lambda.

        Returns:
            Lambda FileSystem using the root access point.
        """
        return self.file_system.as_lambda_file_system(self.root_access_point)


@dataclass
class MountPointConfiguration:
    """Configuration for mounting an EFS file system.

    Attributes:
        file_system (Optional[Union[efs.FileSystem, efs.IFileSystem]]):
            The EFS file system to mount.
        access_point (Optional[Union[efs.AccessPoint, efs.IAccessPoint]]):
            The access point to use for mounting.
        mount_point (str): The path where the file system will be mounted.
        root_directory (Optional[str]): Root directory within the file system.
        read_only (bool): Whether to mount as read-only. Defaults to False.

    Raises:
        ValueError: If neither file system nor access point is provided,
            or if access point's file system doesn't match the provided file system.
    """

    file_system: efs.FileSystem | efs.IFileSystem | None
    access_point: efs.AccessPoint | efs.IAccessPoint | None
    mount_point: str
    root_directory: str | None = None
    read_only: bool = False

    def __post_init__(self):
        if not self.access_point and not self.file_system:
            raise ValueError("Must provide either file system or access point")
        if (
            self.access_point
            and self.file_system
            and self.access_point.file_system.file_system_id != self.file_system.file_system_id
        ):
            raise ValueError("File system of Access point and file system must be the same")
        if not self.mount_point.startswith("/"):
            raise ValueError("Mount point must start with /")

    @classmethod
    def from_file_system(
        cls,
        file_system: efs.FileSystem | efs.IFileSystem,
        root_directory: str | None = None,
        mount_point: str | None = None,
        read_only: bool = False,
    ) -> "MountPointConfiguration":
        """Create configuration from a file system.

        Args:
            file_system (Union[efs.FileSystem, efs.IFileSystem]): The file system.
            root_directory (Optional[str]): Root directory. Defaults to "/".
            mount_point (Optional[str]): Mount point path.
                Defaults to /opt/efs/{file_system_id}.
            read_only (bool): Mount as read-only. Defaults to False.

        Returns:
            MountPointConfiguration for the file system.
        """
        if not root_directory:
            root_directory = "/"
        if not mount_point:
            mount_point = f"/opt/efs/{file_system.file_system_id}"
        return cls(
            mount_point=mount_point,
            file_system=file_system,
            access_point=None,
            root_directory=root_directory,
            read_only=read_only,
        )

    @classmethod
    def from_access_point(
        cls,
        access_point: efs.AccessPoint | efs.IAccessPoint,
        mount_point: str | None = None,
        read_only: bool = False,
    ) -> "MountPointConfiguration":
        """Create configuration from an access point.

        Args:
            access_point (Union[efs.AccessPoint, efs.IAccessPoint]): The access point.
            mount_point (Optional[str]): Mount point path.
                Defaults to /opt/efs/{access_point_id}.
            read_only (bool): Mount as read-only. Defaults to False.

        Returns:
            MountPointConfiguration for the access point.
        """
        if not mount_point:
            mount_point = f"/opt/efs/{access_point.access_point_id}"
        return cls(
            mount_point=mount_point,
            access_point=access_point,
            file_system=None,
            root_directory=None,
            read_only=read_only,
        )

    @property
    def file_system_id(self) -> str:
        """Get the file system ID.

        Returns:
            The EFS file system ID.

        Raises:
            ValueError: If no file system or access point is configured.
        """
        if self.access_point:
            return self.access_point.file_system.file_system_id
        elif self.file_system:
            return self.file_system.file_system_id
        else:
            raise ValueError("No file system or access point provided")

    @property
    def access_point_id(self) -> str | None:
        """Get the access point ID.

        Returns:
            The access point ID, or None if using file system directly.
        """
        if self.access_point:
            return self.access_point.access_point_id
        return None

    def to_batch_mount_point(self, name: str, sfn_format: bool = False) -> dict[str, Any]:
        """Convert to Batch mount point configuration.

        Args:
            name (str): Name of the volume.
            sfn_format (bool): Use Step Functions API case. Defaults to False.

        Returns:
            Dictionary containing the mount point configuration.
        """
        mount_point: dict[str, Any] = to_mount_point(
            self.mount_point, self.read_only, source_volume=name
        )  # type: ignore[arg-type, assignment]  # typed dict should be accepted
        if sfn_format:
            return convert_to_sfn_api_action_case(mount_point)
        return mount_point

    def to_batch_volume(self, name: str, sfn_format: bool = False) -> dict[str, Any]:
        """Convert to Batch volume configuration.

        Args:
            name (str): Name of the volume.
            sfn_format (bool): Use Step Functions API case. Defaults to False.

        Returns:
            Dictionary containing the volume configuration.
        """
        efs_volume_configuration: dict[str, Any] = {
            "fileSystemId": self.file_system_id,
        }
        if self.access_point:
            efs_volume_configuration["transitEncryption"] = "ENABLED"
            # TODO: Consider adding IAM
            efs_volume_configuration["authorizationConfig"] = {
                "accessPointId": self.access_point.access_point_id,
                "iam": "DISABLED",
            }
        else:
            efs_volume_configuration["rootDirectory"] = self.root_directory or "/"
        volume: dict[str, Any] = to_volume(
            None,
            name=name,
            efs_volume_configuration=efs_volume_configuration,  # type: ignore
        )
        if sfn_format:
            return convert_to_sfn_api_action_case(volume)
        return volume


def create_access_point(
    scope: constructs.Construct,
    file_system: efs.FileSystem | efs.IFileSystem,
    name: str,
    path: str,
    *tags: EFSTag | tuple[str, str],
) -> efs.AccessPoint:
    """Create an EFS access point.

    Uses CfnAccessPoint because the AccessPoint construct does not support
    tagging or naming. Tags are used to set the name.

    Args:
        scope (constructs.Construct): The construct scope.
        file_system (Union[efs.FileSystem, efs.IFileSystem]): The file system.
        name (str): Name used for construct ID and as default Name tag.
        path (str): Access point path within the file system.
        *tags (Union[EFSTag, Tuple[str, str]]): Variable number of tags.

    Returns:
        The created access point.
    """
    ap_tags = [tag if isinstance(tag, EFSTag) else EFSTag(*tag) for tag in tags]

    if not any(tag.key == "Name" for tag in ap_tags):
        ap_tags.insert(0, EFSTag("Name", name))

    cfn_access_point = efs.CfnAccessPoint(
        scope,
        f"{name}-cfn-ap",
        file_system_id=file_system.file_system_id,
        access_point_tags=[
            efs.CfnAccessPoint.AccessPointTagProperty(key=tag.key, value=tag.value)
            for tag in ap_tags
        ],
        posix_user=efs.CfnAccessPoint.PosixUserProperty(
            gid="0",
            uid="0",
        ),
        root_directory=efs.CfnAccessPoint.RootDirectoryProperty(
            creation_info=efs.CfnAccessPoint.CreationInfoProperty(
                owner_gid="0",
                owner_uid="0",
                permissions="0777",
            ),
            path=path,
        ),
    )
    return efs.AccessPoint.from_access_point_attributes(
        scope,
        f"{name}-access-point",
        access_point_id=cfn_access_point.attr_access_point_id,
        file_system=file_system,
    )  # type: ignore


def grant_connectable_file_system_access(
    file_system: efs.IFileSystem | efs.FileSystem,
    connectable: ec2.IConnectable,
    permissions: Literal["r", "rw"] = "rw",
) -> None:
    """Grant a connectable resource access to an EFS file system.

    Args:
        file_system (Union[efs.IFileSystem, efs.FileSystem]): The file system.
        connectable (ec2.IConnectable): The connectable resource.
        permissions (Literal["r", "rw"]): Permission level. Defaults to "rw".
    """
    file_system.connections.allow_default_port_from(connectable)
    repair_connectable_efs_dependency(file_system, connectable)


def grant_role_file_system_access(
    file_system: efs.IFileSystem | efs.FileSystem,
    role: iam.IRole | None,
    permissions: Literal["r", "rw"] = "rw",
) -> None:
    """Grant an IAM role access to an EFS file system.

    Args:
        file_system (Union[efs.IFileSystem, efs.FileSystem]): The file system.
        role (Optional[iam.IRole]): The IAM role to grant access to.
        permissions (Literal["r", "rw"]): Permission level. Defaults to "rw".
    """
    grant_managed_policies(role, "AmazonElasticFileSystemReadOnlyAccess")
    if "w" in permissions:
        grant_managed_policies(role, "AmazonElasticFileSystemClientReadWriteAccess")


def grant_grantable_file_system_access(
    file_system: efs.IFileSystem | efs.FileSystem,
    grantable: iam.IGrantable,
    permissions: Literal["r", "rw"] = "rw",
) -> None:
    """Grant a grantable principal access to an EFS file system.

    Args:
        file_system (Union[efs.IFileSystem, efs.FileSystem]): The file system.
        grantable (iam.IGrantable): The grantable principal.
        permissions (Literal["r", "rw"]): Permission level. Defaults to "rw".
    """
    actions = []
    if "w" in permissions:
        actions.append("elasticfilesystem:ClientWrite")
    file_system.grant(grantable, *actions)


def grant_file_system_access(
    file_system: efs.IFileSystem | efs.FileSystem, resource: lambda_.Function
) -> None:
    """Grant a Lambda function full access to an EFS file system.

    Grants grantable, role, and connectable access.

    Args:
        file_system (Union[efs.IFileSystem, efs.FileSystem]): The file system.
        resource (lambda_.Function): The Lambda function.
    """
    grant_grantable_file_system_access(file_system, resource)
    grant_role_file_system_access(file_system, resource.role)
    grant_connectable_file_system_access(file_system, resource)


def repair_connectable_efs_dependency(
    file_system: efs.IFileSystem | efs.FileSystem, connectable: ec2.IConnectable
) -> None:
    """Repair cyclical dependency between EFS and dependent connectable.

    When an EFS filesystem is added to a Lambda Function (via the file_system= param)
    it automatically sets up networking access between the two by adding an ingress
    rule on the EFS security group. However, the ingress rule resource gets attached
    to whichever CDK Stack the EFS security group is defined on.

    If the Lambda Function is defined on a different stack, it creates a circular
    dependency issue, where the EFS stack is dependent on the Lambda security group's
    ID and the Lambda stack is dependent on the EFS stack's file system object.

    To resolve this, we manually remove the ingress rule that gets automatically
    created and recreate it on the Lambda's stack instead.

    Based on: https://github.com/aws/aws-cdk/issues/18759#issuecomment-1268689132

    Args:
        file_system (Union[efs.IFileSystem, efs.FileSystem]): The EFS file system.
        connectable (ec2.IConnectable): The connectable resource.

    Raises:
        RuntimeError: If unable to remove the child node.
    """
    connections = connectable.connections
    # Collect IDs of all security groups attached to the connections
    connection_sgs = {sg.security_group_id for sg in connections.security_groups}
    # Iterate over the security groups attached to EFS
    for efs_sg in file_system.connections.security_groups:
        # Iterate over the security group's child nodes
        for child in efs_sg.node.find_all():
            # If this is an ingress rule with a "source" equal to one of
            # the connections' security groups
            if (
                isinstance(child, ec2.CfnSecurityGroupIngress)
                and child.source_security_group_id in connection_sgs
            ):
                # Try to remove the node (raise an error if removal fails)
                node_id = child.node.id
                if not efs_sg.node.try_remove_child(node_id):
                    raise RuntimeError(f"Could not remove child node: {node_id}")

    # Finally, configure the connection between the connections object
    # and the EFS file system which will define the new ingress rule on
    # the stack defining the connection object instead.
    connections.allow_to_default_port(file_system)



__all__ = [
    "get_package_root",
    "resolve_repo_root",
    "get_env_base",
    "get_config",
]

import logging
import os
import pathlib
from typing import cast

import constructs
from aibs_informatics_core.env import (
    ENV_BASE_KEY_ALIAS,
    ENV_LABEL_KEY_ALIAS,
    ENV_TYPE_KEY_ALIAS,
    LABEL_KEY,
    LABEL_KEY_ALIAS,
    EnvBase,
    EnvType,
)
from aibs_informatics_core.utils.os_operations import get_env_var, set_env_var

from aibs_informatics_cdk_lib.project.config import BaseProjectConfig, G, P, ProjectConfig, S

logger = logging.getLogger(__name__)


def resolve_repo_root(start_path: str | pathlib.Path | None = None) -> str:
    """Find the root of the git repository

    Returns:
        Absolute root path
    """
    return str(
        next(
            filter(
                lambda p: (p / ".git").is_dir(),  # type: ignore
                pathlib.Path(start_path or __file__).absolute().parents,
            )
        )
    )


def get_package_root() -> str:
    """Find the root package

    ASSUMPTION: the infrastructure package name is "aibs-informatics-cdk-lib"

    Returns:
        Absolute root path
    """
    return str(
        next(
            filter(
                lambda p: (p / "aibs-informatics-cdk-lib").is_dir(),  # type: ignore
                pathlib.Path(__file__).absolute().parents,
            )
        )
    )


ENV_BASE_KEYS = [EnvBase.ENV_BASE_KEY, ENV_BASE_KEY_ALIAS, "env"]
ENV_TYPE_KEYS = [EnvBase.ENV_TYPE_KEY, ENV_TYPE_KEY_ALIAS]
ENV_LABEL_KEYS = [EnvBase.ENV_LABEL_KEY, ENV_LABEL_KEY_ALIAS, LABEL_KEY, LABEL_KEY_ALIAS]


def get_env_base(node: constructs.Node) -> EnvBase:
    """Resolves EnvBase from cdk context or environment

    Order of resolution:
        1. CDK context value (specifying -c K=V)
            1. look for EnvBase
            2. look for EnvType/EnvLabel
        2. env variable
            1. Look for EnvBase
            2. Look for EnvType/EnvLabel

    Args:
        node (constructs.Node): cdk construct node

    Returns:
        Environment base
    """
    # We need to check from context node FIRST before reading from env variables
    env_base_str = _get_from_context(node, ENV_BASE_KEYS)
    env_type_str = _get_from_context(node, ENV_TYPE_KEYS)
    env_label__from_context = _get_from_context(node, ENV_LABEL_KEYS)
    env_label__from_env = EnvBase.load_env_label__from_env()
    if env_base_str:
        env_base = EnvBase(env_base_str)
        logger.info(f"Loading EnvBase from CDK CONTEXT: env_base={env_base}")
        return env_base

    if env_type_str:
        env_type = EnvType(env_type_str)
        env_label = env_label__from_context or env_label__from_env
        env_base = EnvBase.from_type_and_label(env_type=env_type, env_label=env_label)
        logger.info(f"Loading EnvBase from type/label CDK CONTEXT: env_base={env_base}")
        return env_base

    # Now we try to load env base from environment
    try:
        env_base = EnvBase.from_env()
        logger.info(f"Loading EnvBase from ENV_VARs: env_base={env_base}")
        return env_base
    except Exception:
        env_type = EnvType.DEV

        env_label = env_label__from_context
        if env_label is None:
            if env_label__from_env:
                env_label = env_label__from_env
            else:
                env_label = (get_env_var("USER") or "anonymous").split(".")[0]
        logger.info(
            f"No EnvBase defaults set by CDK CONTEXT or ENV VAR. "
            f"Using env_type={env_type}, env_label={env_label}"
        )
        return EnvBase.from_type_and_label(env_type=env_type, env_label=env_label)


def set_env_base(env_base: EnvBase) -> None:
    """Set the environment base

    Args:
        env_base (EnvBase): environment base
    """
    set_env_var(EnvBase.ENV_BASE_KEY, env_base)
    set_env_var(EnvBase.ENV_TYPE_KEY, env_base.env_type)
    if env_base.env_label:
        set_env_var(EnvBase.ENV_LABEL_KEY, env_base.env_label)
    else:
        os.environ.pop(EnvBase.ENV_LABEL_KEY, None)


def get_project_config_and_env_base(
    node: constructs.Node, project_config_cls: type[P] | None = None
) -> tuple[P, EnvBase]:
    """Get the project configuration and environment base for a given node."""
    if project_config_cls is None:
        project_config_cls = cast(type[P], ProjectConfig)
    env_base = get_env_base(node)

    config = project_config_cls.load_config()
    return config, env_base


def get_config(
    node: constructs.Node,
    project_config_cls: type[BaseProjectConfig[G, S]] = ProjectConfig,  # type: ignore # the type signature is a bit complex and mypy is having trouble with it, but the intention is that the caller can specify a project config class with any generic params and the return type will be the stage config type of that project config class
) -> S:
    """
    Retrieves the stage configuration for a given node.

    Args:
        node (constructs.Node): The node for which to retrieve the configuration.
        project_config_cls (Type[BaseProjectConfig[G, S]], optional):
            The project configuration class to use. Defaults to ProjectConfig.

    Returns:
        The stage configuration object.

    """
    project_config, env_base = get_project_config_and_env_base(  # type: ignore
        node, project_config_cls=project_config_cls
    )
    set_env_base(env_base)

    stage_config: S = project_config.get_stage_config(env_type=env_base.env_type)
    stage_config.env.label = env_base.env_label

    return stage_config


def _get_from_context(
    node: constructs.Node, keys: list[str], default: str | None = None
) -> str | None:
    for key in keys:
        value = node.try_get_context(key)
        if value is not None:
            return value
    else:
        return default



"""Compute service constructs for AWS Batch.

This module provides high-level constructs for creating Batch compute
environments with various configurations.
"""

from abc import abstractmethod
from collections.abc import Iterable

from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3
from constructs import Construct

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.batch.infrastructure import (
    Batch,
    BatchEnvironment,
    BatchEnvironmentConfig,
)
from aibs_informatics_cdk_lib.constructs_.batch.instance_types import (
    LAMBDA_LARGE_INSTANCE_TYPES,
    LAMBDA_MEDIUM_INSTANCE_TYPES,
    LAMBDA_SMALL_INSTANCE_TYPES,
    ON_DEMAND_INSTANCE_TYPES,
    SPOT_INSTANCE_TYPES,
)
from aibs_informatics_cdk_lib.constructs_.batch.launch_template import BatchLaunchTemplateBuilder
from aibs_informatics_cdk_lib.constructs_.batch.types import BatchEnvironmentDescriptor
from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration


class BaseBatchComputeConstruct(EnvBaseConstruct):
    """Base class for Batch compute constructs.

    Abstract base class that provides common functionality for creating
    and managing AWS Batch compute environments.

    """

    def __init__(
        self,
        scope: Construct,
        id: str | None,
        env_base: EnvBase,
        vpc: ec2.Vpc,
        batch_name: str,
        buckets: Iterable[s3.Bucket] | None = None,
        file_systems: Iterable[efs.FileSystem | efs.IFileSystem] | None = None,
        mount_point_configs: Iterable[MountPointConfiguration] | None = None,
        instance_role_name: str | None = None,
        instance_role_policy_statements: list[iam.PolicyStatement] | None = None,
        **kwargs,
    ) -> None:
        """Initialize a Batch compute construct.

        Args:
            scope (Construct): The construct scope.
            id (Optional[str]): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
            vpc (ec2.Vpc): VPC for the compute environments.
            batch_name (str): Name for the batch infrastructure.
            buckets (Optional[Iterable[s3.Bucket]]): S3 buckets to grant access to.
            file_systems (Optional[Iterable[Union[efs.FileSystem, efs.IFileSystem]]]):
                EFS file systems to grant access to.
            mount_point_configs (Optional[Iterable[MountPointConfiguration]]):
                Mount point configurations for EFS.
            instance_role_name (Optional[str]): Name for the instance role.
            instance_role_policy_statements (Optional[List[iam.PolicyStatement]]):
                Additional IAM policy statements.
            **kwargs: Additional arguments passed to parent.
        """
        super().__init__(scope, id, env_base, **kwargs)
        self.batch_name = batch_name
        self.batch = Batch(
            self,
            batch_name,
            self.env_base,
            vpc=vpc,
            instance_role_name=instance_role_name,
            instance_role_policy_statements=instance_role_policy_statements,
        )

        self.create_batch_environments()

        bucket_list = list(buckets or [])

        file_system_list = list(file_systems or [])

        if mount_point_configs:
            mount_point_config_list = list(mount_point_configs)
            file_system_list = self._update_file_systems_from_mount_point_configs(
                file_system_list, mount_point_config_list
            )
        else:
            mount_point_config_list = self._get_mount_point_configs(file_system_list)

        # Validation to ensure that the file systems are not duplicated
        self._validate_mount_point_configs(mount_point_config_list)

        self.grant_storage_access(*bucket_list, *file_system_list)

    @property
    @abstractmethod
    def primary_batch_environment(self) -> BatchEnvironment:
        """Get the primary batch environment.

        Returns:
            The primary BatchEnvironment for this compute construct.
        """
        raise NotImplementedError()

    @abstractmethod
    def create_batch_environments(self) -> None:
        """Create the batch environments.

        Subclasses must implement this to create their specific
        batch environment configurations.
        """
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """Get the batch name.

        Returns:
            The batch infrastructure name.
        """
        return self.batch_name

    def grant_storage_access(
        self, *resources: s3.Bucket | efs.FileSystem | efs.IFileSystem
    ) -> None:
        """Grant access to storage resources.

        Args:
            *resources (Union[s3.Bucket, efs.FileSystem, efs.IFileSystem]):
                Variable number of storage resources to grant access to.
        """
        self.batch.grant_instance_role_permissions(read_write_resources=list(resources))

        for batch_environment in self.batch.environments:
            for resource in resources:
                if isinstance(resource, efs.FileSystem):
                    batch_environment.grant_file_system_access(resource)

    def _validate_mount_point_configs(
        self, mount_point_configs: list[MountPointConfiguration]
    ) -> None:
        """Validate mount point configurations for duplicates.

        Args:
            mount_point_configs (List[MountPointConfiguration]): Configs to validate.

        Raises:
            ValueError: If duplicate mount points are found.
        """
        _ = {}
        for mpc in mount_point_configs:
            if mpc.mount_point in _ and _[mpc.mount_point] != mpc:
                raise ValueError(
                    f"Mount point {mpc.mount_point} is duplicated. "
                    "Cannot have multiple mount points configurations with the same name."
                )
            _[mpc.mount_point] = mpc

    def _get_mount_point_configs(
        self, file_systems: list[efs.FileSystem | efs.IFileSystem] | None
    ) -> list[MountPointConfiguration]:
        """Get mount point configurations from file systems.

        Args:
            file_systems (Optional[List[Union[efs.FileSystem, efs.IFileSystem]]]):
                File systems to create mount configs for.

        Returns:
            List of MountPointConfiguration objects.
        """
        mount_point_configs = []
        if file_systems:
            for fs in file_systems:
                mount_point_configs.append(MountPointConfiguration.from_file_system(fs))
        return mount_point_configs

    def _update_file_systems_from_mount_point_configs(
        self,
        file_systems: list[efs.FileSystem | efs.IFileSystem],
        mount_point_configs: list[MountPointConfiguration],
    ) -> list[efs.FileSystem | efs.IFileSystem]:
        """Update file systems list from mount point configurations.

        Args:
            file_systems (List[Union[efs.FileSystem, efs.IFileSystem]]): Existing file systems.
            mount_point_configs (List[MountPointConfiguration]): Mount configs to process.

        Returns:
            Updated list of file systems.

        Raises:
            ValueError: If mount config has neither file system nor access point.
        """
        file_system_map: dict[str, efs.FileSystem | efs.IFileSystem] = {
            fs.file_system_id: fs for fs in file_systems
        }
        for mpc in mount_point_configs:
            if mpc.file_system_id not in file_system_map:
                if not mpc.file_system and mpc.access_point:
                    file_system_map[mpc.file_system_id] = mpc.access_point.file_system
                elif mpc.file_system:
                    file_system_map[mpc.file_system_id] = mpc.file_system
                else:
                    raise ValueError(
                        "Mount point configuration must have a file system or access point."
                    )

        return list(file_system_map.values())


class BatchCompute(BaseBatchComputeConstruct):
    """Standard Batch compute construct with on-demand, spot, and Fargate environments.

    Provides a complete Batch compute setup with three environment types:
    on-demand, spot, and Fargate.

    Attributes:
        on_demand_batch_environment: On-demand compute environment.
        spot_batch_environment: Spot compute environment.
        fargate_batch_environment: Fargate compute environment.
    """

    @property
    def primary_batch_environment(self) -> BatchEnvironment:
        """Get the primary batch environment.

        Returns:
            The on-demand batch environment.
        """
        return self.on_demand_batch_environment

    def create_batch_environments(self) -> None:
        """Create on-demand, spot, and Fargate batch environments."""
        lt_builder = BatchLaunchTemplateBuilder(
            self, f"{self.name}-lt-builder", env_base=self.env_base
        )
        self.on_demand_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-on-demand"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.BEST_FIT,
                instance_types=[ec2.InstanceType(_) for _ in ON_DEMAND_INSTANCE_TYPES],
                use_spot=False,
                use_fargate=False,
                use_public_subnets=False,
            ),
            launch_template_builder=lt_builder,
        )

        self.spot_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-spot"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.SPOT_PRICE_CAPACITY_OPTIMIZED,
                instance_types=[ec2.InstanceType(_) for _ in SPOT_INSTANCE_TYPES],
                use_spot=True,
                use_fargate=False,
                use_public_subnets=False,
            ),
            launch_template_builder=lt_builder,
        )

        self.fargate_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-fargate"),
            config=BatchEnvironmentConfig(
                allocation_strategy=None,
                instance_types=None,
                use_spot=False,
                use_fargate=True,
                use_public_subnets=False,
            ),
            launch_template_builder=lt_builder,
        )


class LambdaCompute(BatchCompute):
    """Lambda-optimized Batch compute construct.

    Provides Batch environments optimized for Lambda-like workloads with
    small, medium, and large instance type configurations.

    """

    @property
    def primary_batch_environment(self) -> BatchEnvironment:
        """Get the primary batch environment.

        Returns:
            The main Lambda batch environment.
        """
        return self.lambda_batch_environment

    def create_batch_environments(self) -> None:
        """Create Lambda-optimized batch environments."""
        lt_builder = BatchLaunchTemplateBuilder(
            self, f"{self.name}-lt-builder", env_base=self.env_base
        )
        self.lambda_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-lambda"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.BEST_FIT,
                instance_types=[
                    *LAMBDA_SMALL_INSTANCE_TYPES,
                    *LAMBDA_MEDIUM_INSTANCE_TYPES,
                    *LAMBDA_LARGE_INSTANCE_TYPES,
                ],
                use_spot=False,
                use_fargate=False,
                use_public_subnets=False,
                minv_cpus=2,
            ),
            launch_template_builder=lt_builder,
        )

        self.lambda_small_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-lambda-small"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.BEST_FIT,
                instance_types=[*LAMBDA_SMALL_INSTANCE_TYPES],
                use_spot=False,
                use_fargate=False,
                use_public_subnets=False,
            ),
            launch_template_builder=lt_builder,
        )

        self.lambda_medium_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-lambda-medium"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.BEST_FIT,
                instance_types=[*LAMBDA_MEDIUM_INSTANCE_TYPES],
                use_spot=False,
                use_fargate=False,
                use_public_subnets=False,
                minv_cpus=2,
            ),
            launch_template_builder=lt_builder,
        )

        self.lambda_large_batch_environment = self.batch.setup_batch_environment(
            descriptor=BatchEnvironmentDescriptor(f"{self.name}-lambda-large"),
            config=BatchEnvironmentConfig(
                allocation_strategy=batch.AllocationStrategy.BEST_FIT,
                instance_types=[*LAMBDA_LARGE_INSTANCE_TYPES],
                use_spot=False,
                use_fargate=False,
                use_public_subnets=False,
            ),
            launch_template_builder=lt_builder,
        )



#!/usr/bin/env python3

import aws_cdk as cdk
from constructs import Construct

from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.project.config import StageConfig
from aibs_informatics_cdk_lib.project.utils import get_config
from aibs_informatics_cdk_lib.stages.base import ConfigBasedStage
from aibs_informatics_core_app.stacks.assets import AIBSInformaticsAssetsStack
from aibs_informatics_core_app.stacks.core import CoreStack
from aibs_informatics_core_app.stacks.demand_execution import (
    DemandExecutionInfrastructureStack,
    DemandExecutionStack,
)


class InfraStage(ConfigBasedStage):
    def __init__(self, scope: Construct, config: StageConfig, **kwargs) -> None:
        super().__init__(scope, "Infra", config, **kwargs)
        assets = AIBSInformaticsAssetsStack(
            self,
            self.get_stack_name("Assets"),
            self.env_base,
            env=self.env,
        )
        core = CoreStack(
            self,
            self.get_stack_name("Core"),
            self.env_base,
            name="core",
            env=self.env,
        )

        demand_execution_infra = DemandExecutionInfrastructureStack(
            self,
            self.get_stack_name("DemandExecutionInfra"),
            self.env_base,
            vpc=core.vpc,
            buckets=[core.bucket],
            mount_point_configs=[
                MountPointConfiguration.from_file_system(core.file_system, None, "/opt/efs"),
            ],
            env=self.env,
        )

        demand_execution = DemandExecutionStack(  # noqa: F841
            self,
            self.get_stack_name("DemandExecution"),
            env_base=self.env_base,
            assets=assets.assets,
            scaffolding_bucket=core.bucket,
            efs_ecosystem=core.efs_ecosystem,
            data_sync_job_queue=demand_execution_infra.infra_compute.lambda_medium_batch_environment.job_queue_name,
            scaffolding_job_queue=demand_execution_infra.infra_compute.primary_batch_environment.job_queue_name,
            execution_job_queue=demand_execution_infra.execution_compute.primary_batch_environment.job_queue_name,
            env=self.env,
        )


def main():
    app = cdk.App()

    config: StageConfig = get_config(app.node)

    InfraStage(app, config)

    app.synth()


if __name__ == "__main__":
    main()



from typing import Any

import constructs
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.constructs_.sfn.utils import convert_reference_paths_in_mapping


class S3Operation:
    @classmethod
    def put_object(
        cls,
        scope: constructs.Construct,
        id: str,
        bucket_name: str,
        key: str,
        body: Any,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        """Create a chain to put a body of text to S3

        This chain consists of two states:
            1. Pass state (resolving references and resctructuring inputs)
            2. API Call to put object.

        The context at the end state of the chain should contain the following fields:
            - Bucket
            - Key

        All parameters can be either a reference (e.g. $.path.to.my.value)
        or an explicit value.

        Examples:
            Context:
                {}
            Definition:
                S3Operation.put_object(..., bucket_name="bucket", key="key", body="body")
            Result:
                # text "body" is put at s3://bucket/key
                {"Bucket": "bucket", "Key": "key"}

            Context:
                {"bucket": "woah", "key": "wait/what", "nested": {"a": "b"}}
            Definition:
                S3Operation.put_object(..., bucket_name="$.bucket", key="$.key", body="$.nested")
            Result:
                # text '{"a": "b"}' is put at s3://woah/wait/what
                {"Bucket": "woah", "Key": "wait/what"}

        Args:
            scope (constructs.Construct): scope construct
            id (str): An ID prefix
            bucket_name (str): explicit or reference to name of bucket
            key (str): explicit or reference to name of key
            body (Any): explicit or reference to body to upload

        Returns:
            sfn.Chain
        """
        init = sfn.Pass(
            scope,
            id + " PutObject Prep",
            parameters=convert_reference_paths_in_mapping(
                {
                    "Bucket": bucket_name,
                    "Key": key,
                    "Body": body,
                }
            ),
            result_path=result_path or "$",
        )

        state_json = {
            "Type": "Task",
            "Resource": "arn:aws:states:::aws-sdk:s3:putObject",
            "Parameters": {
                "Bucket.$": f"{result_path or '$'}.Bucket",
                "Key.$": f"{result_path or '$'}.Key",
                "Body.$": f"{result_path or '$'}.Body",
            },
            # "ResultSelector": {
            #     "ETag.$": "$.Bucket",
            #     "Key.$": "$.Key",
            # },
            "ResultPath": sfn.JsonPath.DISCARD,
            "Retry": [
                {
                    "ErrorEquals": ["S3.S3Exception"],
                    # Interval at attempt n = IntervalSeconds x BackoffRate ^(n-1)
                    # Total time from first try: 3 + 6 + 12 + 24 + 48 = 93 seconds
                    "IntervalSeconds": 3,
                    "MaxAttempts": 5,
                    "BackoffRate": 2.0,
                    "JitterStrategy": "FULL",
                },
            ],
        }
        put_object = sfn.CustomState(scope, id + " PutObject API Call", state_json=state_json)

        end = sfn.Pass(
            scope,
            id + " PutObject Post",
            parameters=convert_reference_paths_in_mapping(
                {
                    "Bucket": f"{result_path or '$'}.Bucket",
                    "Key": f"{result_path or '$'}.Key",
                }
            ),
            result_path=result_path,
            output_path=output_path,
        )

        return sfn.Chain.start(init).next(put_object).next(end)

    @classmethod
    def get_object(
        cls,
        scope: constructs.Construct,
        id: str,
        bucket_name: str,
        key: str,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        """Creates a chain to get a body of text from S3

        This chain consists of two states:
            1. Pass state (resolving references and resctructuring inputs)
            2. API Call to get object.

        The context at the end state of the chain should contain the following fields:
            - Body
            - Bucket
            - Key

        All parameters can be either a reference (e.g. $.path.to.my.value)
        or an explicit value.

        Examples:
            Context:
                {}
            Definition:
                S3Operation.get_object(..., bucket_name="bucket", key="key")
            Result:
                # text "body" is fetched from s3://bucket/key
                {"Body": "body"}

            Context:
                {"bucket": "woah", "key": "wait/what"}
            Definition:
                S3Operation.get_object(..., bucket_name="$.bucket", key="$.key")
            Result:
                # text '{"a": "b"}' is fetched from s3://woah/wait/what
                {"Body": '{"a": "b"}'}

        Args:
            scope (constructs.Construct): scope construct
            id (str): An ID prefix
            bucket_name (str): explicit or reference to name of bucket
            key (str): explicit or reference to name of key

        Returns:
            sfn.Chain
        """
        init = sfn.Pass(
            scope,
            id + " GetObject Prep",
            parameters=convert_reference_paths_in_mapping(
                {
                    "Bucket": bucket_name,
                    "Key": key,
                }
            ),
            result_path=result_path or "$",
        )

        if result_path:
            bucket_path = f"{result_path}.Bucket"
            key_path = f"{result_path}.Key"
        else:
            bucket_path = "$.Bucket"
            key_path = "$.Key"

        state_json = {
            "Type": "Task",
            "Resource": "arn:aws:states:::aws-sdk:s3:getObject",
            "Parameters": {
                "Bucket.$": bucket_path,
                "Key.$": key_path,
            },
            "ResultSelector": {
                "Body.$": "$.Body",
            },
            "ResultPath": result_path,
            "OutputPath": output_path,
            "Retry": [
                {
                    "ErrorEquals": ["S3.S3Exception"],
                    # Interval at attempt n = IntervalSeconds x BackoffRate ^(n-1)
                    # Total time from first try: 3 + 6 + 12 + 24 + 48 = 93 seconds
                    "IntervalSeconds": 3,
                    "MaxAttempts": 5,
                    "BackoffRate": 2.0,
                    "JitterStrategy": "FULL",
                },
            ],
        }

        get_object = sfn.CustomState(scope, id + " GetObject API Call", state_json=state_json)

        return sfn.Chain.start(init).next(get_object)

    @classmethod
    def put_payload(
        cls,
        scope: constructs.Construct,
        id: str,
        payload: str,
        bucket_name: str,
        key: str | None = None,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        """Puts a payload to s3 and returns the location of the payload in s3

        All parameters can be either a reference (e.g. $.path.to.my.value)
        or an explicit value.

        Examples:
            Context:
                {"a": "b"}
            Definition:
                S3Operation.put_payload(..., bucket_name="bucket", payload="$")
            Result:
                # text '{"a": "b"}' is written to s3://bucket/1234.../...1234
                {"Bucket": "bucket", "Key": "1234.../...1234"}

            Context:
                {"bucket": "woah", "key": "wait/what", "data": {"a": "b"}}
            Definition:
                S3Operation.put_payload(..., bucket_name="$.bucket", payload="$.data")
            Result:
                # text '{"a": "b"}' is written to s3://woah/wait/what
                {"Bucket": "woah", "Key": "wait/what"}


        Args:
            scope (constructs.Construct): cdk construct
            id (str): id
            payload (str): explicit value or reference path in the context object
                (e.g. "$", "$.path")
            bucket_name (str): explicit value or reference path for bucket name
            key (Optional[str], optional): explicit value or reference path for key.
                If not provided, the following Key is generated:
                    {$$.Execution.Name}/{UUID}

        Returns:
            sfn.Chain
        """
        key = key or sfn.JsonPath.format("{}/{}", sfn.JsonPath.execution_name, sfn.JsonPath.uuid())

        put_chain = S3Operation.put_object(
            scope, id, bucket_name, key, payload, result_path, output_path
        )
        return put_chain

    @classmethod
    def get_payload(
        cls,
        scope: constructs.Construct,
        id: str,
        bucket_name: str,
        key: str,
        result_path: str | None = "$",
    ) -> sfn.Chain:
        """Gets a payload from s3

        This chain fetches object and then passes the body through a json parser

        The resulting payload will be stored to path specified by result_path

        Examples:
            Use Case #1 - No result path
                Context:
                    {"bucket": "woah", "key": "wait/what"}
                Definition:
                    S3Operation.get_payload(..., bucket_name="$.bucket", key="$.key")
                Result:
                    # text '{"a": "b"}' is fetched from s3://woah/wait/what
                    {"a": "b"}

            Use Case #2 - result path specified
                Context:
                    {"bucket": "woah", "key": "wait/what"}
                Definition:
                    S3Operation.get_payload(
                        ..., bucket_name="$.bucket", key="$.key", result_path="$.payload"
                    )
                Result:
                    # text '{"a": "b"}' is fetched from s3://woah/wait/what
                    {
                        "bucket": "woah",
                        "key": "wait/what",
                        "payload": {"a": "b"}
                    }

        Args:
            scope (constructs.Construct): cdk construct
            id (str): id
            bucket_name (str): bucket name. Can be a reference path
            key (str): key name. Can be a reference path
            result_path (Optional[str], optional): path to store the payload. Defaults to "$".

        Returns:
            sfn.Chain: chain
        """

        get_chain = S3Operation.get_object(scope, id, bucket_name, key, result_path)
        post = sfn.Pass(
            scope,
            id + " Post",
            parameters={
                "Payload": sfn.JsonPath.string_to_json(
                    sfn.JsonPath.string_at(f"{result_path}.Body")
                )
            },
            result_path=result_path,
        )
        restructure = sfn.Pass(
            scope,
            id + " Restructure",
            input_path=f"{result_path}.Payload",
            result_path=result_path,
        )

        return get_chain.next(post).next(restructure)



# flake8: noqa

import base64
import json
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, List, Optional, TypeVar

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_cloudwatch as cw
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_iam as iam

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.batch.types import IBatchEnvironmentDescriptor
from aibs_informatics_cdk_lib.constructs_.cw import (
    AlarmMetricConfig,
    GraphMetricConfig,
    GroupedGraphMetricConfig,
)
from aibs_informatics_cdk_lib.constructs_.s3 import EnvBaseBucket

T = TypeVar("T", bound="BatchLaunchTemplateUserData")


class IBatchLaunchTemplateBuilder(EnvBaseConstruct, Generic[T]):
    def grant_instance_role_permissions(self, instance_role: iam.Role):
        pass

    @abstractmethod
    def create_launch_template(
        self,
        descriptor: IBatchEnvironmentDescriptor,
        security_group: ec2.SecurityGroup | None = None,
        **kwargs,
    ) -> ec2.LaunchTemplate:
        raise NotImplementedError()


class BatchLaunchTemplateBuilder(IBatchLaunchTemplateBuilder["BatchLaunchTemplateUserData"]):
    def grant_instance_role_permissions(self, instance_role: iam.Role):
        pass

    def create_launch_template(
        self,
        descriptor: IBatchEnvironmentDescriptor,
        security_group: ec2.SecurityGroup | None = None,
        python_version: str = "python3.11",
        **kwargs,
    ) -> ec2.LaunchTemplate:
        user_data = ec2.UserData.custom(
            BatchLaunchTemplateUserData(
                env_base=self.env_base,
                batch_env_name=descriptor.get_name(),
                python_version=python_version,
            ).user_data_text
        )

        launch_template_name = f"{self.env_base}-{descriptor.get_name()}-launch-template"

        launch_template = ec2.LaunchTemplate(
            self,
            launch_template_name,
            # NOTE: unsetting because of complications when multiple batch environments are created
            # launch_template_name=launch_template_name,
            instance_initiated_shutdown_behavior=ec2.InstanceInitiatedShutdownBehavior.TERMINATE,
            security_group=security_group,
            user_data=user_data,
        )
        return launch_template


class EbsBatchLaunchTemplateBuilder(IBatchLaunchTemplateBuilder["EbsBatchLaunchTemplateUserData"]):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        assets_bucket: EnvBaseBucket,
    ) -> None:
        super().__init__(scope, id, env_base)

        self.assets_bucket = assets_bucket

        self.docker_volume_device_name = "/dev/xvdba"

    def grant_instance_role_permissions(self, instance_role: iam.Role):
        pass

    def create_launch_template(
        self,
        descriptor: IBatchEnvironmentDescriptor,
        security_group: ec2.SecurityGroup | None = None,
        python_version: str = "python3.11",
        **kwargs,
    ) -> ec2.LaunchTemplate:
        user_data = ec2.UserData.custom(
            EbsBatchLaunchTemplateUserData(
                env_base=self.env_base,
                batch_env_name=descriptor.get_name(),
                docker_volume_device_name=self.docker_volume_device_name,
                python_version=python_version,
            ).user_data_text
        )

        launch_template_name = f"{self.env_base}-{descriptor}-launch-template"
        launch_template = ec2.LaunchTemplate(
            self,
            launch_template_name,
            launch_template_name=launch_template_name,
            block_devices=[
                ec2.BlockDevice(
                    device_name=self.docker_volume_device_name,
                    volume=ec2.BlockDeviceVolume.ebs(
                        30,
                        encrypted=True,
                        volume_type=ec2.EbsDeviceVolumeType.GP3,
                        iops=3000,
                    ),
                ),
            ],
            user_data=user_data,
            instance_initiated_shutdown_behavior=ec2.InstanceInitiatedShutdownBehavior.TERMINATE,
            security_group=security_group,
        )
        return launch_template


@dataclass
class BatchLaunchTemplateUserData:
    env_base: EnvBase
    batch_env_name: str
    python_version: str
    user_data_text: str = field(init=False)

    def __post_init__(self):
        self.user_data_text = DEFAULT_LAUNCH_TEMPLATE_USER_DATA.format(
            python_version=self.python_version,
            config_json=self.config_builder.to_string(),
        )

    @property
    def config_builder(self) -> "CloudWatchConfigBuilder":
        return CloudWatchConfigBuilder(self.env_base, self.batch_env_name)

    def get_base64_user_data(self) -> str:
        return base64.b64encode(self.user_data_text.encode("ascii")).decode("ascii")


@dataclass
class EbsBatchLaunchTemplateUserData(BatchLaunchTemplateUserData):
    docker_volume_device_name: str
    python_version: str

    def __post_init__(self):
        self.user_data_text = EBS_AUTOSCALE_LAUNCH_TEMPLATE_USER_DATA.format(
            device_name=self.docker_volume_device_name,
            config_json=self.config_builder.to_string(),
            python_version=self.python_version,
        )


@dataclass
class CloudWatchConfigBuilder:
    env_base: EnvBase
    batch_env_name: str

    @property
    def metric_namespace(self) -> str:
        return self.env_base.get_metric_namespace("CWAgent")

    def to_json(self) -> dict[str, Any]:
        """Builds a CW Agent config

        # https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/CloudWatch-Agent-Configuration-File-Details.html

        Returns:
            Dict[str, Any]: config
        """
        return {
            "agent": {
                "metrics_collection_interval": 60,
                "logfile": "/opt/aws/amazon-cloudwatch-agent/logs/amazon-cloudwatch-agent.log",
            },
            "logs": self.get_logs_config(),
            "metrics": self.get_metrics_config(),
        }

    def to_string(self) -> str:
        config = self.to_json()
        return base64.b64encode(json.dumps(config).encode("ascii")).decode("ascii")

    def get_logs_config(self) -> dict[str, Any]:
        """Generates CW Agent sub config for logs

        https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/CloudWatch-Agent-Configuration-File-Details.html#CloudWatch-Agent-Configuration-File-Logssection

        Returns:
            Dict[str, Any]: CW logs config
        """
        env_base = self.env_base
        return {
            "logs_collected": {
                "files": {
                    "collect_list": [
                        {
                            "file_path": "/opt/aws/amazon-cloudwatch-agent/logs/amazon-cloudwatch-agent.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/amazon-cloudwatch-agent.log",
                        },
                        {
                            "file_path": "/var/log/cloud-init.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/cloud-init.log",
                        },
                        {
                            "file_path": "/var/log/cloud-init-output.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/cloud-init-output.log",
                        },
                        {
                            "file_path": "/var/log/ecs/ecs-init.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/ecs-init.log",
                        },
                        {
                            "file_path": "/var/log/ecs/ecs-agent.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/ecs-agent.log",
                        },
                        {
                            "file_path": "/var/log/ecs/ecs-volume-plugin.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/ecs-volume-plugin.log",
                        },
                        {
                            "file_path": "/var/log/ebs-autoscale-install.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/ebs-autoscale-install.log",
                        },
                        {
                            "file_path": "/var/log/ebs-autoscale.log",
                            "log_group_name": f"/aws/ecs/container-instance/{env_base}",
                            "log_stream_name": f"/aws/ecs/container-instance/{env_base}/{{instance_id}}/ebs-autoscale.log",
                        },
                    ]
                }
            }
        }

    def get_metrics_config(self) -> dict[str, Any]:
        """Generates metrics config section of CW Agent config

        https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/CloudWatch-Agent-Configuration-File-Details.html#CloudWatch-Agent-Configuration-File-Metricssection

        Returns:
            Dict[str, Any]: metrics config
        """
        return {
            "namespace": self.metric_namespace,
            "metrics_collected": self._get_metrics_collected(),
            "append_dimensions": {
                "ImageId": "${aws:ImageId}",
                "InstanceId": "${aws:InstanceId}",
                "InstanceType": "${aws:InstanceType}",
                "AutoScalingGroupName": "${aws:AutoScalingGroupName}",
            },
            "aggregation_dimensions": [
                ["AutoScalingGroupName"],
                ["InstanceId", "InstanceType"],
                ["env_base", "batch_env_name"],
                [],
            ],
        }

    def _get_metrics_collected(self) -> dict[str, Any]:
        metrics_collected: dict[str, Any] = {}
        # https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/metrics-collected-by-CloudWatch-agent.html

        DEFAULT_COLLECTION_INTERVAL = 60 * 1
        DEFAULT_APPEND_DIMENSIONS = {
            "env_base": self.env_base,
            "batch_env_name": self.batch_env_name,
        }

        metrics_collected["cpu"] = {
            "measurement": ["cpu_usage_active"],
            "resources": ["*"],
            "metrics_collection_interval": DEFAULT_COLLECTION_INTERVAL,
            "append_dimensions": DEFAULT_APPEND_DIMENSIONS,
        }

        metrics_collected["disk"] = {
            "measurement": [
                "disk_used_percent",
            ],
            "resources": ["*"],
            "metrics_collection_interval": DEFAULT_COLLECTION_INTERVAL,
            "append_dimensions": DEFAULT_APPEND_DIMENSIONS,
        }

        metrics_collected["mem"] = {
            "measurement": [
                "mem_used_percent",
                "mem_used",
                "mem_free",
            ],
            "metrics_collection_interval": DEFAULT_COLLECTION_INTERVAL,
            "append_dimensions": DEFAULT_APPEND_DIMENSIONS,
        }
        metrics_collected["net"] = {
            "measurement": [
                "net_bytes_recv",
                "net_bytes_sent",
            ],
            "metrics_collection_interval": DEFAULT_COLLECTION_INTERVAL,
            "append_dimensions": DEFAULT_APPEND_DIMENSIONS,
        }

        return metrics_collected

    def get_grouped_graph_metric_configs(self) -> list[GroupedGraphMetricConfig]:
        PERC_ALARM_METRIC_FN = lambda name, threshold: AlarmMetricConfig(
            name=f"{self.batch_env_name}-{name}",
            threshold=threshold,
            evaluation_periods=3,
            datapoints_to_alarm=3,
            comparison_operator=cw.ComparisonOperator.GREATER_THAN_THRESHOLD,
        )
        graph_metric_configs = [
            GroupedGraphMetricConfig(
                title="CPU Usage",
                namespace=self.metric_namespace,
                metrics=[
                    GraphMetricConfig(
                        metric="cpu_usage_active",
                        statistic="AVG",
                        alarm=PERC_ALARM_METRIC_FN("cpu_usage_active", 95),
                    )
                ],
            ),
            GroupedGraphMetricConfig(
                title="Memory Usage",
                namespace=self.metric_namespace,
                metrics=[
                    GraphMetricConfig(
                        metric="mem_used_percent",
                        statistic="AVG",
                        alarm=PERC_ALARM_METRIC_FN("mem_used_percent", 95),
                    )
                ],
            ),
            GroupedGraphMetricConfig(
                title="Network Traffic",
                namespace=self.metric_namespace,
                metrics=[
                    GraphMetricConfig(metric="net_bytes_recv", statistic="SUM"),
                    GraphMetricConfig(metric="net_bytes_sent", statistic="SUM"),
                ],
            ),
        ]
        return graph_metric_configs


DEFAULT_LAUNCH_TEMPLATE_USER_DATA = """MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="==BOUNDARY=="

--==BOUNDARY==
Content-Type: text/cloud-config; charset="us-ascii"

#cloud-config
repo_update: true
repo_upgrade: security
apt:
  sources:
    deadsnakes-ppa-source:
      source: "ppa:deadsnakes/ppa"

packages:
- jq
- btrfs-progs
- sed
- git
- {python_version}
- {python_version}-venv
- amazon-efs-utils
- amazon-ssm-agent
- unzip
- amazon-cloudwatch-agent

write_files:
- permissions: "0644"
  encoding: b64
  path: /opt/aws/amazon-cloudwatch-agent/etc/config.json
  content: {config_json}

runcmd:
- systemctl start amazon-ssm-agent

# Start aws cloudwatch agent
- /opt/aws/amazon-cloudwatch-agent/bin/amazon-cloudwatch-agent-ctl -a fetch-config -m ec2 -s -c file:/opt/aws/amazon-cloudwatch-agent/etc/config.json

# enable ecs spot instance draining
- echo ECS_ENABLE_SPOT_INSTANCE_DRAINING=true >> /etc/ecs/ecs.config

# pull docker images
- echo ECS_IMAGE_PULL_BEHAVIOR=default >> /etc/ecs/ecs.config

# Enable ECS Metadata file
- echo ECS_ENABLE_CONTAINER_METADATA=true >> /etc/ecs/ecs.config

# Install lustre
- amazon-linux-extras install lustre -y

--==BOUNDARY==--"""


EBS_AUTOSCALE_LAUNCH_TEMPLATE_USER_DATA = """MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="==BOUNDARY=="

--==BOUNDARY==
Content-Type: text/cloud-config; charset="us-ascii"

#cloud-config
repo_update: true
repo_upgrade: security
apt:
  sources:
    deadsnakes-ppa-source:
      source: "ppa:deadsnakes/ppa"

packages:
- jq
- btrfs-progs
- sed
- git
- {python_version}
- {python_version}-venv
- amazon-efs-utils
- amazon-ssm-agent
- unzip
- amazon-cloudwatch-agent

write_files:
- permissions: "0644"
  encoding: b64
  path: /opt/aws/amazon-cloudwatch-agent/etc/config.json
  content: {config_json}

runcmd:
- systemctl start amazon-ssm-agent

# Start aws cloudwatch agent
- /opt/aws/amazon-cloudwatch-agent/bin/amazon-cloudwatch-agent-ctl -a fetch-config -m ec2 -s -c file:/opt/aws/amazon-cloudwatch-agent/etc/config.json

# install aws-cli v2 and copy the static binary in an easy to find location for bind-mounts into containers
- mkdir -p /opt/aws-cli/bin
- curl -s --fail --retry 3 --retry-connrefused "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "/tmp/awscliv2.zip" && unzip -q /tmp/awscliv2.zip -d /tmp && /tmp/aws/install -b /usr/bin && cp -a -f $(dirname $(find /usr/local/aws-cli -name 'aws' -type f))/. /opt/aws-cli/bin/
- command -v aws || sleep 5 && curl -s --fail --retry 3 --retry-connrefused "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "/tmp/awscliv2.zip" && unzip -q /tmp/awscliv2.zip -d /tmp && /tmp/aws/install -b /usr/bin && cp -a -f $(dirname $(find /usr/local/aws-cli -name 'aws' -type f))/. /opt/aws-cli/bin/
- command -v aws || sleep 10 && curl -s --fail --retry 3 --retry-connrefused "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "/tmp/awscliv2.zip" && unzip -q /tmp/awscliv2.zip -d /tmp && /tmp/aws/install -b /usr/bin && cp -a -f $(dirname $(find /usr/local/aws-cli -name 'aws' -type f))/. /opt/aws-cli/bin/
- command -v aws || echo "Unable to install AWS CLI v2"

## Enable EBS Autoscale
- EBS_AUTOSCALE_VERSION=$(curl --silent "https://api.github.com/repos/awslabs/amazon-ebs-autoscale/releases/latest" | jq -r .tag_name)
- cd /opt && git clone https://github.com/awslabs/amazon-ebs-autoscale.git
- cd /opt/amazon-ebs-autoscale && git checkout $EBS_AUTOSCALE_VERSION
- cp -au /var/lib/docker /var/lib/docker.bk
- rm -rf /var/lib/docker/*
- sh /opt/amazon-ebs-autoscale/install.sh -d {device_name} -f btrfs --volume-type gp3 --volume-iops 10000 --initial-utilization-threshold 65 --max-ebs-volume-size 500 -m /var/lib/docker 2>&1 > /var/log/ebs-autoscale-install.log
- docker_storage_options="DOCKER_STORAGE_OPTIONS=\"--storage-driver btrfs\""
- awk -v docker_storage_options="$docker_storage_options" \
'{{ sub(/DOCKER_STORAGE_OPTIONS=.*/, docker_storage_options); print }}' \
/etc/sysconfig/docker-storage \
> /opt/amazon-ebs-autoscale/docker-storage
- mv -f /opt/amazon-ebs-autoscale/docker-storage /etc/sysconfig/docker-storage
- cp -au /var/lib/docker.bk/* /var/lib/docker

# enable ecs spot instance draining
- echo ECS_ENABLE_SPOT_INSTANCE_DRAINING=true >> /etc/ecs/ecs.config

# pull docker images
- echo ECS_IMAGE_PULL_BEHAVIOR=default >> /etc/ecs/ecs.config

# Enable ECS Metadata file
- echo ECS_ENABLE_CONTAINER_METADATA=true >> /etc/ecs/ecs.config

--==BOUNDARY==--"""



"""Base stack classes for environment-aware CDK stacks.

This module provides the foundational stack classes that enable
environment-aware deployments with automatic tagging and resource naming.
"""

__all__ = [
    "EnvBaseStack",
    "EnvBaseStackMixins",
    "add_stack_dependencies",
    "get_all_stacks",
]

from typing import cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase, EnvType

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


def get_all_stacks(scope: constructs.Construct) -> list[cdk.Stack]:
    """Get all CDK stacks from a construct scope.

    Args:
        scope (constructs.Construct): The construct scope to search.

    Returns:
        List of all Stack children in the scope.
    """
    children = scope.node.children
    return [cast(cdk.Stack, child) for child in children if isinstance(child, cdk.Stack)]


def add_stack_dependencies(source_stack: cdk.Stack, dependent_stacks: list[cdk.Stack]):
    """Add dependencies between stacks.

    Makes the dependent stacks depend on the source stack,
    ensuring proper deployment order.

    Args:
        source_stack (cdk.Stack): The stack that others depend on.
        dependent_stacks (List[cdk.Stack]): Stacks that depend on the source.
    """
    for dependent_stack in dependent_stacks:
        dependent_stack.add_dependency(source_stack)


class EnvBaseStackMixins(EnvBaseConstructMixins):
    """Mixin class for environment-aware stack functionality.

    Inherits all functionality from EnvBaseConstructMixins for use in stacks.
    """

    pass


class EnvBaseStack(cdk.Stack, EnvBaseStackMixins):
    """Base stack class with environment awareness.

    Provides automatic tagging with environment information and
    environment-specific removal policies.

    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        env: cdk.Environment | None = None,
        **kwargs,
    ) -> None:
        """Initialize an environment-aware stack.

        Args:
            scope (constructs.Construct): The parent construct scope.
            id (Optional[str]): The stack ID. Auto-generated if None.
            env_base (EnvBase): The environment base configuration.
            env (Optional[cdk.Environment]): AWS environment settings.
            **kwargs: Additional arguments passed to cdk.Stack.
        """
        super().__init__(
            scope,
            id or env_base.get_construct_id(str(self.__class__)),
            env=env,
            **kwargs,
        )
        self.env_base = env_base
        self.add_tags(*self.stack_tags)

    @property
    def aws_region(self) -> str:
        return cdk.Stack.of(self).region

    @property
    def aws_account(self) -> str:
        return cdk.Stack.of(self).account

    @property
    def stack_tags(self) -> list[cdk.Tag]:
        return [
            *self.construct_tags,
            cdk.Tag(key=self.env_base.ENV_BASE_KEY, value=self.env_base),
        ]

    @property
    def removal_policy(self) -> cdk.RemovalPolicy:
        if self.env_base.env_type == EnvType.DEV:
            return cdk.RemovalPolicy.DESTROY
        return cdk.RemovalPolicy.RETAIN



from dataclasses import dataclass
from typing import Literal

from attr import field
from aws_cdk import aws_cloudwatch as cw
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.constructs_.cw.types import AlarmMetricConfig, GraphMetricConfig

SFN_TIME_UNITS = Literal["hours", "minutes", "seconds", "milliseconds"]


@dataclass
class StateMachineMetricConfigGenerator:
    state_machine: sfn.IStateMachine
    state_machine_name: str
    dimension_map: dict = field(init=False)

    def __post_init__(self):
        self.dimension_map = {"StateMachineArn": self.state_machine.state_machine_arn}

    def get_execution_completion_metric(
        self, name_override: str | None = None
    ) -> GraphMetricConfig:
        """get the execution completion metric for the state machine

        Args:
            name_override (Optional[str], optional): override for name used.
                Defaults to None.

        Returns:
            GraphMetricConfig
        """
        return GraphMetricConfig(
            metric="ExecutionsSucceeded",
            label=f"{name_override or self.state_machine_name} Completed",
            statistic="Sum",
            dimension_map=self.dimension_map,
        )

    def get_execution_invocations_metric(
        self, name_override: str | None = None
    ) -> GraphMetricConfig:
        """get the execution invocations metric for the state machine

        Args:
            name_override (Optional[str], optional): override for name used.
                Defaults to None.

        Returns:
            GraphMetricConfig
        """
        return GraphMetricConfig(
            metric="ExecutionsStarted",
            label=f"{name_override or self.state_machine_name} Started",
            statistic="Sum",
            dimension_map=self.dimension_map,
        )

    def get_execution_failures_metric(
        self,
        name_override: str | None = None,
        discriminator: str | None = None,
        alarm_threshold: int = 1,
        alarm_evaluation_periods: int = 3,
        alarm_datapoints_to_alarm: int = 1,
    ) -> GraphMetricConfig:
        """get the execution failures metric for the state machine

        Args:
            name_override (Optional[str], optional): override for name used.
                Defaults to state machine name.
            discriminator (Optional[str], optional): Required if grouping with other metric configs that specify the same metric math.
                Defaults to "0".
            alarm_threshold (int, optional): Alarm threshold used. Defaults to 1.
            alarm_evaluation_periods (int, optional): Alarm evaluation periods. Defaults to 3.
            alarm_datapoints_to_alarm (int, optional): Alarm datapoints to alarm. Defaults to 1.

        Returns:
            Graph metric config for execution failures
        """  # noqa: E501
        name = name_override or self.state_machine_name
        idx = discriminator or "0"
        return GraphMetricConfig(
            metric="ExecutionErrors",
            statistic="Sum",
            label=f"{name} Errors",
            dimension_map=self.dimension_map,
            metric_expression=(
                f"failed_{idx} + aborted_{idx} + timed_out_{idx} + throttled_{idx}"
            ),
            using_metrics={
                f"failed_{idx}": self.state_machine.metric_failed(),
                f"aborted_{idx}": self.state_machine.metric_aborted(),
                f"timed_out_{idx}": self.state_machine.metric_timed_out(),
                f"throttled_{idx}": self.state_machine.metric_throttled(),
            },
            alarm=AlarmMetricConfig(
                name=f"{name}-errors",
                threshold=alarm_threshold,
                evaluation_periods=alarm_evaluation_periods,
                datapoints_to_alarm=alarm_datapoints_to_alarm,
                comparison_operator=cw.ComparisonOperator.GREATER_THAN_THRESHOLD,
            ),
        )

    def get_execution_timing_metric(
        self,
        name_override: str | None = None,
        discriminator: str | None = None,
        time_unit: SFN_TIME_UNITS = "minutes",
    ) -> GraphMetricConfig:
        """get the execution time metric for the state machine

        Args:
            name_override (Optional[str], optional): override for name used.
                Defaults to state machine name.
            discriminator (Optional[str], optional): Required if grouping with other metric configs
                that specify the same metric math. Defaults to "0".
            time_unit (SFN_TIME_UNITS, optional): unit of time to use for metric.
                Defaults to "minutes".

        Returns:
            GraphMetricConfig
        """
        name = name_override or self.state_machine_name
        idx = discriminator or "0"
        if time_unit == "seconds":
            divisor = " / 1000"
        elif time_unit == "minutes":
            divisor = " / 1000 / 60"
        elif time_unit == "hours":
            divisor = " / 1000 / 60 / 60"
        else:
            divisor = ""

        return GraphMetricConfig(
            metric="ExecutionTime",
            statistic="Average",
            label=f"{name} Execution Time",
            dimension_map=self.dimension_map,
            metric_expression=f"time_msec_{idx} {divisor}",
            using_metrics={f"time_msec_{idx}": self.state_machine.metric_time()},
        )



import constructs
from aibs_informatics_core.env import EnvBase

from aibs_informatics_cdk_lib.constructs_.assets.code_asset_definitions import (
    AIBSInformaticsAssets,
)
from aibs_informatics_cdk_lib.stacks.base import EnvBaseStack


class AIBSInformaticsAssetsStack(EnvBaseStack):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        **kwargs,
    ):
        super().__init__(scope, id, env_base, **kwargs)

        self.assets = AIBSInformaticsAssets(
            self,
            "aibs-info-assets",
            self.env_base,
        )
        self.code_assets = self.assets.code_assets
        self.docker_assets = self.assets.docker_assets



__all__ = [
    "AlarmMetricConfig",
    "EnhancedDashboard",
    "DashboardTools",
    "GraphMetricConfig",
    "GroupedGraphMetricConfig",
]


from .dashboard import DashboardTools, EnhancedDashboard
from .types import AlarmMetricConfig, GraphMetricConfig, GroupedGraphMetricConfig



import re
from collections.abc import Mapping
from functools import reduce
from re import Pattern
from typing import Any, ClassVar, TypeVar, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.tools.dicttools import convert_key_case
from aibs_informatics_core.utils.tools.strtools import pascalcase
from aws_cdk import aws_stepfunctions as sfn

T = TypeVar("T")


def convert_reference_paths_in_mapping(parameters: Mapping[str, Any]) -> Mapping[str, Any]:
    return {k: convert_reference_paths(v) for k, v in parameters.items()}


def convert_to_sfn_api_action_case(parameters: T) -> T:
    """Converts a dictionary of parameters to the format expected by the Step Functions for service integration.

    Even if a native API specifies a parameter in camelCase, the Step Functions SDK expects it in pascal case.

    https://docs.aws.amazon.com/step-functions/latest/dg/supported-services-awssdk.html#use-awssdk-integ

    Args:
        parameters (Dict[str, Any]): parameters for SDK action

    Returns:
        Dict[str, Any]: parameters for SDK action in pascal case
    """  # noqa: E501
    return convert_key_case(parameters, pascalcase)


def convert_reference_paths(parameters: JSON) -> JSON:
    if isinstance(parameters, dict):
        return {k: convert_reference_paths(v) for k, v in parameters.items()}
    elif isinstance(parameters, list):
        return [convert_reference_paths(v) for v in parameters]
    elif isinstance(parameters, str):
        if (
            parameters.startswith("$") or parameters.startswith("States.")
        ) and not parameters.startswith("${Token"):
            return sfn.JsonPath.string_at(parameters)
        else:
            return parameters
    else:
        return parameters


def enclosed_chain(
    scope: constructs.Construct,
    id: str,
    definition: sfn.IChainable,
    input_path: str | None = None,
    result_path: str | None = None,
) -> sfn.Chain:
    """Enclose the current state machine fragment within a parallel state.

    Notes:
        - If input_path is not provided, it will default to "$"
        - If result_path is not provided, it will default to input_path

    Args:
        id (str): an identifier for the parallel state
        input_path (Optional[str], optional): input path for the enclosed state.
            Defaults to "$".
        result_path (Optional[str], optional): result path to put output of enclosed state.
            Defaults to same as input_path.

    Returns:
        sfn.Chain: the new state machine fragment
    """
    if input_path is None:
        input_path = "$"
    if result_path is None:
        result_path = input_path

    chain = (
        sfn.Chain.start(definition)
        if not isinstance(definition, (sfn.Chain, sfn.StateMachineFragment))
        else definition
    )

    if isinstance(chain, sfn.Chain):
        parallel = chain.to_single_state(
            id=f"{id} Enclosure", input_path=input_path, result_path=result_path
        )
    else:
        parallel = chain.to_single_state(input_path=input_path, result_path=result_path)
    definition = sfn.Chain.start(parallel)

    if result_path and result_path != sfn.JsonPath.DISCARD:
        restructure = sfn.Pass(
            scope,
            f"{id} Enclosure Post",
            input_path=f"{result_path}[0]",
            result_path=result_path,
        )
        definition = definition.next(restructure)

    return definition


class JsonReferencePath(str):
    """
    str extension with properties that provide *some* functionality for defining JsonPath reference expressions
    More details: https://github.com/json-path/JsonPath

    Primarily supports "$" reference.


    """  # noqa: E501

    _EXTRA_PERIODS_PATTERN: ClassVar[Pattern[str]] = re.compile(r"[$.]+")
    _PERIOD_PATTERN: ClassVar[Pattern[str]] = re.compile(r"(?<!\$)\.(?!\$)")
    _PREFIX: ClassVar[str] = "$."
    _SUFFIX: ClassVar[str] = ".$"
    _REF: ClassVar[str] = "$"

    def __new__(cls, content: str):
        return super().__new__(cls, cls.sanitize(content))

    def __add__(self, other):
        return JsonReferencePath(super().__add__("." + other))

    def extend(self, *paths: str) -> "JsonReferencePath":
        return cast(JsonReferencePath, reduce(lambda x, y: x + y, [self, *paths]))

    @property
    def as_key(self) -> str:
        """
        Returns reference path as a key. This appends ".$"
        """
        return f"{self}{self._SUFFIX}" if self else self._REF

    @property
    def as_reference(self) -> str:
        """
        Returns reference path as a value. This prepends "$."
        """
        return f"{self._PREFIX}{self}" if self else self._REF

    @property
    def as_jsonpath_string(self) -> str:
        return sfn.JsonPath.string_at(self.as_reference)

    @property
    def as_jsonpath_object(self) -> cdk.IResolvable:
        return sfn.JsonPath.object_at(self.as_reference)

    @property
    def as_jsonpath_json_to_string(self) -> str:
        return sfn.JsonPath.json_to_string(self.as_jsonpath_object)

    @property
    def as_jsonpath_list(self) -> list[str]:
        return sfn.JsonPath.list_at(self.as_reference)

    @property
    def as_jsonpath_number(self) -> int | float:
        return sfn.JsonPath.number_at(self.as_reference)

    @classmethod
    def sanitize(cls, s: str) -> str:
        """Sanitizes a string to ensure string has non-consecutive periods and not on the edge."""
        return f"{cls._EXTRA_PERIODS_PATTERN.sub('.', s).strip('.')}"

    @classmethod
    def is_reference(cls, s: Any) -> bool:
        return isinstance(s, JsonReferencePath) or isinstance(s, str) and s.startswith("$")

    @classmethod
    def empty(cls) -> "JsonReferencePath":
        return cls("")



from typing import Any

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ecr_assets as ecr_assets
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3
from aws_cdk import aws_stepfunctions as sfn
from aws_cdk import aws_stepfunctions_tasks as sfn_tasks

from aibs_informatics_cdk_lib.common.aws.iam_utils import (
    SFN_STATES_EXECUTION_ACTIONS,
    SFN_STATES_READ_ACCESS_ACTIONS,
    batch_policy_statement,
    s3_policy_statement,
    sfn_policy_statement,
)
from aibs_informatics_cdk_lib.common.aws.sfn_utils import JsonReferencePath
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.base import EnvBaseStateMachineFragment
from aibs_informatics_cdk_lib.constructs_.sfn.states.common import CommonOperation


class DemandExecutionFragment(EnvBaseStateMachineFragment, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
        scaffolding_bucket: s3.Bucket,
        scaffolding_job_queue: batch.JobQueue | str,
        batch_invoked_lambda_state_machine: sfn.StateMachine,
        data_sync_state_machine: sfn.StateMachine,
        shared_mount_point_config: MountPointConfiguration | None,
        scratch_mount_point_config: MountPointConfiguration | None,
        tmp_mount_point_config: MountPointConfiguration | None = None,
        context_manager_configuration: dict[str, Any] | None = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        super().__init__(scope, id, env_base)

        # ----------------- Validation -----------------
        if not (shared_mount_point_config and scratch_mount_point_config) or not (
            shared_mount_point_config or scratch_mount_point_config
        ):
            raise ValueError(
                "If shared or scratch mount point configurations are provided,"
                "Both shared and scratch mount point configurations must be provided."
            )

        # ------------------- Setup -------------------

        config_scaffolding_path = "config.scaffolding"
        config_setup_results_path = f"{config_scaffolding_path}.setup_results"
        config_batch_args_path = f"{config_setup_results_path}.batch_args"

        config_cleanup_results_path = "tasks.cleanup.cleanup_results"

        # Create common kwargs for the batch invoked lambda functions
        # - specify the bucket name and job queue
        # - specify the mount points and volumes if provided
        batch_invoked_lambda_kwargs: dict[str, Any] = {
            "bucket_name": scaffolding_bucket.bucket_name,
            "image": aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri,
            "job_queue": scaffolding_job_queue
            if isinstance(scaffolding_job_queue, str)
            else scaffolding_job_queue.job_queue_name,
        }

        # Create request input for the demand scaffolding
        file_system_configurations = {}

        # Update arguments with mount points and volumes if provided
        if shared_mount_point_config or scratch_mount_point_config or tmp_mount_point_config:
            mount_points = []
            volumes = []
            if shared_mount_point_config:
                # update file system configurations for scaffolding function
                file_system_configurations["shared"] = {
                    "file_system": shared_mount_point_config.file_system_id,
                    "access_point": shared_mount_point_config.access_point_id,
                    "container_path": shared_mount_point_config.mount_point,
                }
                # add to mount point and volumes list for batch invoked lambda functions
                mount_points.append(
                    shared_mount_point_config.to_batch_mount_point("shared", sfn_format=True)
                )
                volumes.append(
                    shared_mount_point_config.to_batch_volume("shared", sfn_format=True)
                )

            if scratch_mount_point_config:
                # update file system configurations for scaffolding function
                file_system_configurations["scratch"] = {
                    "file_system": scratch_mount_point_config.file_system_id,
                    "access_point": scratch_mount_point_config.access_point_id,
                    "container_path": scratch_mount_point_config.mount_point,
                }
                # add to mount point and volumes list for batch invoked lambda functions
                mount_points.append(
                    scratch_mount_point_config.to_batch_mount_point("scratch", sfn_format=True)
                )
                volumes.append(
                    scratch_mount_point_config.to_batch_volume("scratch", sfn_format=True)
                )
            if tmp_mount_point_config:
                # update file system configurations for scaffolding function
                file_system_configurations["tmp"] = {
                    "file_system": tmp_mount_point_config.file_system_id,
                    "access_point": tmp_mount_point_config.access_point_id,
                    "container_path": tmp_mount_point_config.mount_point,
                }
                # add to mount point and volumes list for batch invoked lambda functions
                mount_points.append(
                    tmp_mount_point_config.to_batch_mount_point("tmp", sfn_format=True)
                )
                volumes.append(tmp_mount_point_config.to_batch_volume("tmp", sfn_format=True))

            batch_invoked_lambda_kwargs["mount_points"] = mount_points
            batch_invoked_lambda_kwargs["volumes"] = volumes

        request = {
            "demand_execution": sfn.JsonPath.object_at("$"),
            "file_system_configurations": file_system_configurations,
        }
        if context_manager_configuration:
            request["context_manager_configuration"] = context_manager_configuration

        start_state = sfn.Pass(
            self,
            "Start Demand Batch Task",
            parameters={
                "request": request,
            },
        )

        # normalization steps:
        # - merge build and runtime tags

        norm_parallel = CommonOperation.enclose_chainable(
            self,
            "Normalize Demand Execution",
            self.demand_execution_normalize_tags_chain(tags),
            input_path="$.request.demand_execution",
            result_path="$.request.demand_execution",
        )

        prep_scaffolding_task = CommonOperation.enclose_chainable(
            self,
            "Prepare Demand Scaffolding",
            sfn.Pass(
                self,
                "Pass: Prepare Demand Scaffolding",
                parameters={
                    "handler": "aibs_informatics_aws_lambda.handlers.demand.scaffolding.handler",
                    "payload": sfn.JsonPath.object_at("$"),
                    **batch_invoked_lambda_kwargs,
                },
            ).next(
                sfn_tasks.StepFunctionsStartExecution(
                    self,
                    "SM: Prepare Demand Scaffolding",
                    state_machine=batch_invoked_lambda_state_machine,
                    integration_pattern=sfn.IntegrationPattern.RUN_JOB,
                    associate_with_parent=False,
                    input_path="$",
                    output_path="$.Output",
                )
            ),
            input_path="$.request",
            result_path=f"$.{config_scaffolding_path}",
        )

        create_def_and_prepare_job_args_task = CommonOperation.enclose_chainable(
            self,
            "Create Definition and Prep Job Args",
            sfn.Pass(
                self,
                "Pass: Create Definition and Prep Job Args",
                parameters={
                    "handler": "aibs_informatics_aws_lambda.handlers.batch.create.handler",
                    "payload": sfn.JsonPath.object_at("$"),
                    **batch_invoked_lambda_kwargs,
                },
            ).next(
                sfn_tasks.StepFunctionsStartExecution(
                    self,
                    "SM: Create Definition and Prep Job Args",
                    state_machine=batch_invoked_lambda_state_machine,
                    integration_pattern=sfn.IntegrationPattern.RUN_JOB,
                    associate_with_parent=False,
                    input_path="$",
                    output_path="$.Output",
                )
            ),
            input_path="$.batch_create_request",
            result_path="$",
        )

        setup_tasks = (
            sfn.Parallel(
                self,
                "Execution Setup Steps",
                input_path=f"$.{config_scaffolding_path}.setup_configs",
                result_path=f"$.{'.'.join(config_batch_args_path.split('.')[:-1])}",
                result_selector={f"{config_batch_args_path.split('.')[-1]}.$": "$[0]"},
            )
            .branch(create_def_and_prepare_job_args_task)
            .branch(
                sfn.Map(
                    self,
                    "Transfer Inputs TO Batch Job",
                    items_path="$.data_sync_requests",
                ).iterator(
                    sfn_tasks.StepFunctionsStartExecution(
                        self,
                        "Transfer Input",
                        state_machine=data_sync_state_machine,
                        integration_pattern=sfn.IntegrationPattern.RUN_JOB,
                        associate_with_parent=False,
                        result_path=sfn.JsonPath.DISCARD,
                    )
                )
            )
        )

        execution_task = sfn.CustomState(
            self,
            "Submit Batch Job",
            state_json={
                "Type": "Task",
                "Resource": "arn:aws:states:::batch:submitJob.sync",
                "Parameters": {
                    "JobName.$": sfn.JsonPath.string_at(f"$.{config_batch_args_path}.job_name"),
                    "JobDefinition.$": sfn.JsonPath.string_at(
                        f"$.{config_batch_args_path}.job_definition_arn"
                    ),  # noqa: E501
                    "JobQueue.$": sfn.JsonPath.string_at(
                        f"$.{config_batch_args_path}.job_queue_arn"
                    ),  # noqa: E501
                    "Parameters.$": sfn.JsonPath.object_at(
                        f"$.{config_batch_args_path}.parameters"
                    ),  # noqa: E501
                    "ContainerOverrides.$": sfn.JsonPath.object_at(
                        f"$.{config_batch_args_path}.container_overrides"
                    ),  # noqa: E501
                    "Tags.$": sfn.JsonPath.object_at(
                        "$.request.demand_execution.execution_metadata.tags"
                    ),  # noqa: E501
                    "PropagateTags": True,
                },
                "ResultPath": "$.tasks.batch_submit_task",
            },
        )

        cleanup_tasks = sfn.Chain.start(
            sfn.Map(
                self,
                "Transfer Results FROM Batch Job",
                input_path=f"$.{config_scaffolding_path}.cleanup_configs.data_sync_requests",
                result_path=f"$.{config_cleanup_results_path}.transfer_results",
            ).iterator(
                sfn_tasks.StepFunctionsStartExecution(
                    self,
                    "Transfer Result",
                    state_machine=data_sync_state_machine,
                    integration_pattern=sfn.IntegrationPattern.RUN_JOB,
                    associate_with_parent=False,
                    result_path=sfn.JsonPath.DISCARD,
                )
            )
        ).next(
            sfn.Choice(self, "Cleanup Choice")
            .when(
                condition=sfn.Condition.is_present(
                    f"$.{config_scaffolding_path}.cleanup_configs.remove_data_paths_requests"
                ),
                next=sfn.Chain.start(
                    sfn.Map(
                        self,
                        "Map: Cleanup Data Paths",
                        input_path=f"$.{config_scaffolding_path}.cleanup_configs.remove_data_paths_requests",
                        result_path=f"$.{config_cleanup_results_path}.remove_data_paths_results",
                    ).iterator(
                        CommonOperation.enclose_chainable(
                            self,
                            "Cleanup Data Path",
                            definition=sfn.Pass(
                                self,
                                "Pass: Cleanup Data Path",
                                parameters={
                                    "handler": "aibs_informatics_aws_lambda.handlers.data_sync.remove_data_paths_handler",  # noqa: E501
                                    "payload": sfn.JsonPath.object_at("$"),
                                    **batch_invoked_lambda_kwargs,
                                },
                            ).next(
                                sfn_tasks.StepFunctionsStartExecution(
                                    self,
                                    "SM: Cleanup Data Path",
                                    state_machine=batch_invoked_lambda_state_machine,
                                    integration_pattern=sfn.IntegrationPattern.RUN_JOB,
                                    associate_with_parent=False,
                                    input_path="$",
                                    output_path="$.Output",
                                )
                            ),
                        )
                    )
                ),
            )
            .otherwise(sfn.Pass(self, "No Data Paths to Cleanup"))
        )

        # fmt: off
        definition = (
            start_state
            .next(norm_parallel)
            .next(prep_scaffolding_task)
            .next(setup_tasks)
            .next(execution_task)
            .next(cleanup_tasks)
        )
        # fmt: on
        self.definition = definition

    def demand_execution_normalize_tags_chain(self, tags: dict[str, str] | None) -> sfn.IChainable:
        """Merge build and runtime tags.

        Chain Assumptions/Expectations:
            - input is the demand execution request
            - output is the demand execution request with merged tags

        If runtime tags are not provided, only build‑time tags are used.

        Args:
            tags (Optional[Dict[str, str]]): build‑time tags
        Returns:
            sfn.IChainable: the state machine fragment
        """

        static_tags: dict[str, str] = tags or {}  # build‑time default
        execution_tags_path = JsonReferencePath("execution_metadata.tags")

        static_tags = {
            f"{k}.$" if isinstance(v, str) and v.startswith("$") else k: v
            for k, v in static_tags.items()
        }

        merge_tags_chain = CommonOperation.merge_defaults(
            self,
            "Merge Tags",
            input_path="$",
            target_path=execution_tags_path.as_reference,
            defaults=static_tags,
            check_if_target_present=True,
        )
        return merge_tags_chain

    @property
    def required_inline_policy_statements(self) -> list[iam.PolicyStatement]:
        return [
            *super().required_inline_policy_statements,
            batch_policy_statement(self.env_base),
            s3_policy_statement(self.env_base),
            sfn_policy_statement(
                self.env_base,
                actions=SFN_STATES_EXECUTION_ACTIONS + SFN_STATES_READ_ACCESS_ACTIONS,
            ),
            # iam:PassRole needed to support passing batch job role ARNs to batch jobs
            iam.PolicyStatement(
                sid="PassRoleForBatchJobs",
                actions=["iam:PassRole"],
                effect=iam.Effect.ALLOW,
                resources=["*"],
                conditions={"StringLike": {"iam:PassedToService": "ecs-tasks.amazonaws.com"}},
            ),
        ]



from dataclasses import dataclass

from attr import field
from aws_cdk import aws_cloudwatch as cw
from aws_cdk import aws_lambda as lambda_

from aibs_informatics_cdk_lib.constructs_.cw.types import (
    AlarmMetricConfig,
    GraphMetricConfig,
    GroupedGraphMetricConfig,
)


@dataclass
class LambdaFunctionMetricConfigGenerator:
    lambda_function: lambda_.IFunction
    lambda_function_name: str = field(default=None)
    dimension_map: dict = field(init=False)

    def __post_init__(self):
        if self.lambda_function_name is None:
            self.lambda_function_name = self.lambda_function.function_name

        self.dimension_map = {"FunctionName": self.lambda_function_name}

    def get_invocations_metric(
        self,
        name_override: str | None = None,
    ) -> GraphMetricConfig:
        return GraphMetricConfig(
            metric="Invocations",
            label=f"{name_override or self.lambda_function_name} Invocations",
            statistic="Sum",
            dimension_map=self.dimension_map,
        )

    def get_errors_metric(
        self,
        name_override: str | None = None,
        discriminator: str | None = None,
        include_alarm: bool = False,
        alarm_threshold: int = 1,
        alarm_evaluation_periods: int = 3,
        alarm_datapoints_to_alarm: int = 1,
    ) -> GraphMetricConfig:
        name = name_override or self.lambda_function_name
        idx = discriminator or "0"
        config = GraphMetricConfig(
            metric="Errors",
            statistic="Sum",
            label=f"{name} Errors",
            dimension_map=self.dimension_map,
        )
        if include_alarm:
            config["alarm"] = AlarmMetricConfig(
                name=f"{name} Errors Alarm {idx}",
                threshold=alarm_threshold,
                evaluation_periods=alarm_evaluation_periods,
                datapoints_to_alarm=alarm_datapoints_to_alarm,
                comparison_operator=cw.ComparisonOperator.GREATER_THAN_THRESHOLD,
            )
        return config

    def get_availability_metric(
        self,
        name_override: str | None = None,
        discriminator: str | None = None,
    ) -> GraphMetricConfig:
        name = name_override or self.lambda_function_name
        idx = discriminator or "0"

        return GraphMetricConfig(
            metric="Availability",
            statistic="Average",
            dimension_map=self.dimension_map,
            label=f"{name} %",
            metric_expression=f"100 - 100 * errors_{idx} / MAX([errors_{idx}, invocations_{idx}])",
            using_metrics={
                f"errors_{idx}": self.lambda_function.metric_errors(),
                f"invocations_{idx}": self.lambda_function.metric_invocations(),
            },
        )

    def get_duration_avg_metric(
        self,
        name_override: str | None = None,
    ) -> GraphMetricConfig:
        name = name_override or self.lambda_function_name
        return GraphMetricConfig(
            metric="Duration",
            statistic="Average",
            dimension_map=self.dimension_map,
            label=f"{name} Avg",
        )

    def get_duration_max_metric(
        self,
        name_override: str | None = None,
    ) -> GraphMetricConfig:
        name = name_override or self.lambda_function_name
        return GraphMetricConfig(
            metric="Duration",
            statistic="Maximum",
            dimension_map=self.dimension_map,
            label=f"{name} Max",
        )

    def get_duration_min_metric(
        self,
        name_override: str | None = None,
    ) -> GraphMetricConfig:
        name = name_override or self.lambda_function_name
        return GraphMetricConfig(
            metric="Duration",
            statistic="Minimum",
            dimension_map=self.dimension_map,
            label=f"{name} Min",
        )

    def get_duration_metric_group(
        self,
        name_override: str | None = None,
        title: str | None = None,
        include_min_max_duration: bool = False,
    ) -> GroupedGraphMetricConfig:
        name = name_override or self.lambda_function_name

        avg = self.get_duration_avg_metric(name_override)
        if include_min_max_duration:
            min_ = self.get_duration_min_metric(name_override)
            max_ = self.get_duration_max_metric(name_override)

        return GroupedGraphMetricConfig(
            title=title or f"{name} Duration",
            namespace="AWS/Lambda",
            metrics=[avg, min_, max_],
        )

    def get_success_failure_metrics(
        self,
        name_override: str | None = None,
        success_as_percent: bool = True,
    ) -> list[GraphMetricConfig]:
        name = name_override or self.lambda_function_name

        failures = self.get_errors_metric(name)
        if success_as_percent:
            success = self.get_availability_metric(name)
        else:
            success = self.get_invocations_metric(name)
        success["axis_side"] = "right"
        failures["axis_side"] = "left"
        return [success, failures]

    def get_success_failure_metric_group(
        self,
        name_override: str | None = None,
        title: str | None = None,
        success_as_percent: bool = True,
    ) -> GroupedGraphMetricConfig:
        name = name_override or self.lambda_function_name

        failures = self.get_errors_metric(name_override)
        if success_as_percent:
            success = self.get_availability_metric(name_override)
        else:
            success = self.get_invocations_metric(name_override)
        success["axis_side"] = "right"
        failures["axis_side"] = "left"

        return GroupedGraphMetricConfig(
            title=title or f"{name} Invocations",
            namespace="AWS/Lambda",
            metrics=[success, failures],
        )



from aibs_informatics_core.env import EnvBase, ResourceNameBaseEnum
from aibs_informatics_core.models.email_address import EmailAddress
from aibs_informatics_core.utils.hashing import uuid_str
from aws_cdk import Duration
from aws_cdk import aws_cloudwatch as cw
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_sns as sns
from aws_cdk import aws_stepfunctions as sfn
from constructs import Construct

from aibs_informatics_cdk_lib.common.aws.core_utils import build_sfn_arn
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.cw import (
    DashboardTools,
    EnhancedDashboard,
    GraphMetricConfig,
    GroupedGraphMetricConfig,
)
from aibs_informatics_cdk_lib.constructs_.cw.config_generators.lambda_ import (
    LambdaFunctionMetricConfigGenerator,
)
from aibs_informatics_cdk_lib.constructs_.cw.config_generators.sfn import (
    SFN_TIME_UNITS,
    StateMachineMetricConfigGenerator,
)


class MonitoringConstruct(EnvBaseConstruct):
    def __init__(
        self,
        scope: Construct,
        id: str,
        env_base: EnvBase,
        name: str | None = None,
        notify_on_alarms: bool | None = None,
        alarm_topic: sns.Topic | None = None,
    ) -> None:
        super().__init__(scope, id, env_base)
        self.monitoring_name = name or self.construct_id
        self.notify_on_alarms = notify_on_alarms
        self.alarm_topic = alarm_topic

    @property
    def monitoring_name(self) -> str:
        return self._monitoring_name

    @monitoring_name.setter
    def monitoring_name(self, value: str):
        self._monitoring_name = value

    @property
    def notify_on_alarms(self) -> bool:
        if self._notify_on_alarms is None:
            return self.is_test_or_prod
        return self._notify_on_alarms

    @notify_on_alarms.setter
    def notify_on_alarms(self, value: bool | None):
        self._notify_on_alarms = value

    @property
    def alarm_topic(self) -> sns.Topic:
        if self._alarm_topic is None:
            self.alarm_topic = sns.Topic(
                self, self.get_construct_id(self.monitoring_name, "alarm-topic")
            )
            return self.alarm_topic
        else:
            return self._alarm_topic

    @alarm_topic.setter
    def alarm_topic(self, value: sns.Topic | None):
        self._alarm_topic = value

    def create_dashboard(
        self, start: str | None = "-P1W", end: str | None = None
    ) -> EnhancedDashboard:
        return EnhancedDashboard(
            self,
            f"{self.monitoring_name}-dashboard",
            self.env_base,
            dashboard_name=self.get_name_with_env(self.monitoring_name, "Dashboard"),
            start=start,
            end=end,
        )

    def add_function_widget(
        self,
        dashboard: cw.Dashboard,
        function_name: str,
        title: str | None = None,
        title_header_level: int = 1,
        prefix_name_with_env: bool = True,
        include_min_max_duration: bool = False,
    ):
        self.add_function_widgets(
            dashboard,
            function_name,
            title=title,
            title_header_level=title_header_level,
            prefix_name_with_env=prefix_name_with_env,
            include_min_max_duration=include_min_max_duration,
        )

    def add_function_widgets(
        self,
        dashboard: cw.Dashboard,
        *function_names: str,
        title: str | None = None,
        title_header_level: int = 1,
        prefix_name_with_env: bool = True,
        include_min_max_duration: bool = False,
        include_alarm: bool = False,
    ):
        dashboard_tools = (
            dashboard if isinstance(dashboard, EnhancedDashboard) else DashboardTools(dashboard)
        )
        if title:
            dashboard_tools.add_text_widget(title, title_header_level)

        grouped_invocation_metrics: list[GraphMetricConfig] = []
        grouped_error_metrics: list[GraphMetricConfig] = []
        grouped_timing_metrics: list[GraphMetricConfig] = []
        group_name = uuid_str(str(function_names))

        for idx, raw_function_name in enumerate(function_names):
            if prefix_name_with_env:
                function_name = self.get_name_with_env(raw_function_name)
            else:
                function_name = raw_function_name

            fn_config_generator = LambdaFunctionMetricConfigGenerator(
                lambda_function=lambda_.Function.from_function_name(
                    self, f"{function_name}-from-name", function_name=function_name
                ),
                lambda_function_name=function_name,
            )

            grouped_invocation_metrics.append(fn_config_generator.get_invocations_metric())
            grouped_error_metrics.append(
                fn_config_generator.get_errors_metric(
                    discriminator=str(idx), include_alarm=include_alarm
                )
            )

            # Availability metric - make sure to set the axis_side to "right"
            avail_metric = fn_config_generator.get_availability_metric(discriminator=str(idx))
            avail_metric["axis_side"] = "right"
            grouped_error_metrics.append(avail_metric)

            duration_avg_metric = fn_config_generator.get_duration_avg_metric()
            grouped_timing_metrics.append(duration_avg_metric)
            if include_min_max_duration:
                grouped_timing_metrics.append(fn_config_generator.get_duration_max_metric())
                grouped_timing_metrics.append(fn_config_generator.get_duration_min_metric())

        grouped_metrics: list[GroupedGraphMetricConfig] = [
            GroupedGraphMetricConfig(
                title="Function Invocations", metrics=grouped_invocation_metrics
            ),
            GroupedGraphMetricConfig(
                title="Function Successes / Failures", metrics=grouped_error_metrics
            ),
            GroupedGraphMetricConfig(
                title="Function Duration",
                namespace="AWS/Lambda",
                metrics=grouped_timing_metrics,
            ),
        ]

        dashboard_tools.add_graphs(
            grouped_metric_configs=grouped_metrics,
            namespace="AWS/Lambda",
            period=Duration.minutes(5),
            alarm_id_discriminator=group_name,
            alarm_topic=self.alarm_topic if self.notify_on_alarms else None,
            dimensions={},
        )

    def add_state_machine_widget(
        self,
        dashboard: cw.Dashboard,
        state_machine_name: str,
        title: str | None = None,
        title_header_level: int = 1,
        prefix_name_with_env: bool = True,
    ):
        self.add_state_machine_widgets(
            dashboard,
            state_machine_name,
            title=title,
            title_header_level=title_header_level,
            prefix_name_with_env=prefix_name_with_env,
        )

    def add_state_machine_widgets(
        self,
        dashboard: cw.Dashboard,
        *state_machine_names: str,
        title: str | None = None,
        title_header_level: int = 1,
        prefix_name_with_env: bool = True,
        time_unit: SFN_TIME_UNITS = "minutes",
    ):
        dashboard_tools = (
            dashboard if isinstance(dashboard, EnhancedDashboard) else DashboardTools(dashboard)
        )

        if title:
            dashboard_tools.add_text_widget(title, title_header_level)

        grouped_invocation_metrics: list[GraphMetricConfig] = []
        grouped_error_metrics: list[GraphMetricConfig] = []
        grouped_timing_metrics: list[GraphMetricConfig] = []
        group_name = uuid_str(str(state_machine_names))
        for idx, raw_state_machine_name in enumerate(state_machine_names):
            state_machine_name = self.get_state_machine_name(
                raw_state_machine_name, prefix_name_with_env=prefix_name_with_env
            )

            sm_config_generator = StateMachineMetricConfigGenerator(
                state_machine=sfn.StateMachine.from_state_machine_name(
                    self, f"{state_machine_name}-from-name", state_machine_name
                ),
                state_machine_name=state_machine_name,
            )
            grouped_invocation_metrics.append(
                sm_config_generator.get_execution_invocations_metric(raw_state_machine_name)
            )
            grouped_invocation_metrics.append(
                sm_config_generator.get_execution_completion_metric(raw_state_machine_name)
            )
            grouped_error_metrics.append(
                sm_config_generator.get_execution_failures_metric(
                    raw_state_machine_name, discriminator=str(idx)
                )
            )
            grouped_timing_metrics.append(
                sm_config_generator.get_execution_timing_metric(
                    raw_state_machine_name,
                    discriminator=str(idx),
                    time_unit=time_unit,
                )
            )

        grouped_metrics = [
            GroupedGraphMetricConfig(
                title="Execution Invocations", metrics=grouped_invocation_metrics
            ),
            GroupedGraphMetricConfig(title="Execution Errors", metrics=grouped_error_metrics),
            GroupedGraphMetricConfig(
                title="Execution Time",
                metrics=grouped_timing_metrics,
                left_y_axis=cw.YAxisProps(label=f"Time ({time_unit})"),
            ),
        ]

        dashboard_tools.add_graphs(
            grouped_metric_configs=grouped_metrics,
            namespace="AWS/States",
            period=Duration.minutes(5),
            alarm_id_discriminator=group_name,
            alarm_topic=self.alarm_topic if self.notify_on_alarms else None,
            dimensions={},
        )

    def add_alarm_subscription(self, email: str | EmailAddress):
        if not isinstance(email, EmailAddress):
            email = EmailAddress(email)

        return sns.Subscription(
            self,
            self.get_construct_id(f"{email}-alarm-subscription"),
            topic=self.alarm_topic,  # type: ignore  # Topic implements ITopic
            endpoint=email,
            protocol=sns.SubscriptionProtocol.EMAIL,
        )

    def get_state_machine_name(
        self, name: str | ResourceNameBaseEnum, prefix_name_with_env: bool = True
    ) -> str:
        if isinstance(name, ResourceNameBaseEnum):
            return name.get_name(self.env_base)
        elif prefix_name_with_env:
            return self.env_base.get_state_machine_name(name)
        else:
            return name

    def get_state_machine_arn(
        self, name: str | ResourceNameBaseEnum, prefix_name_with_env: bool = True
    ) -> str:
        state_machine_name = self.get_state_machine_name(name, prefix_name_with_env)
        return build_sfn_arn(resource_type="stateMachine", resource_id=state_machine_name)


class ResourceMonitoring(MonitoringConstruct):
    def __init__(
        self,
        scope: Construct,
        id: str,
        env_base: EnvBase,
        notify_on_alarms: bool | None = None,
        alarm_email: str | None = None,
    ) -> None:
        super().__init__(scope, id, env_base)
        self.notify_on_alarms = notify_on_alarms
        if self.notify_on_alarms:
            email = alarm_email or "marmotdev@alleninstitute.org"
            self.add_alarm_subscription(email)

        self.dashboard = self.create_dashboard(start="-P1D")



import aws_cdk as cdk
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2
from constructs import Construct

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.efs.file_system import EFSEcosystem, EnvBaseFileSystem
from aibs_informatics_cdk_lib.constructs_.s3 import EnvBaseBucket, LifecycleRuleGenerator


class Storage(EnvBaseConstruct):
    def __init__(
        self,
        scope: Construct,
        id: str | None,
        env_base: EnvBase,
        name: str,
        vpc: ec2.Vpc,
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.RETAIN,
        **kwargs,
    ) -> None:
        super().__init__(scope, id, env_base, **kwargs)

        self._bucket = EnvBaseBucket(
            self,
            "Bucket",
            self.env_base,
            bucket_name=name,
            removal_policy=removal_policy,
            lifecycle_rules=[
                LifecycleRuleGenerator.expire_files_under_prefix(),
                LifecycleRuleGenerator.expire_files_with_scratch_tags(),
                LifecycleRuleGenerator.use_storage_class_as_default(),
            ],
        )

        self._efs_ecosystem = EFSEcosystem(
            self, id="EFS", env_base=self.env_base, file_system_name=name, vpc=vpc
        )
        self._file_system = self._efs_ecosystem.file_system

    @property
    def bucket(self) -> EnvBaseBucket:
        return self._bucket

    @property
    def efs_ecosystem(self) -> EFSEcosystem:
        return self._efs_ecosystem

    @property
    def file_system(self) -> EnvBaseFileSystem:
        return self._file_system
